from pathlib import Path
from typing import Literal

import brainrender
import numpy as np
import polars as pl
from brainglobe_atlasapi.bg_atlas import BrainGlobeAtlas
from brainrender.actors import Points
from typing_extensions import Self

from argclz import AbstractParser, argument, str_tuple_type, validator
from neuralib.atlas.brainrender.util import get_color
from neuralib.atlas.util import allen_to_brainrender_coord
from neuralib.util.logging import setup_clogger

__all__ = ['BrainRenderCLI']

CAMERA_ANGLE_TYPE = Literal['sagittal', 'sagittal2', 'frontal', 'top', 'top_side', 'three_quarters']
SHADER_STYLE_TYPE = Literal['metallic', 'cartoon', 'plastic', 'shiny', 'glossy']


class BrainRenderCLI(AbstractParser):
    """Reconstruct a 3D brain view used brainrender module"""
    DESCRIPTION = 'Reconstruct a 3D brain view used brainrender module'

    DEFAULT_REGION_COLORS = ['lightblue', 'pink', 'turquoise']

    # ============== #
    # BASIC SETTINGS #
    # ============== #

    GROUP_SETTINGS = 'Basic Settings Option'

    camera_angle: CAMERA_ANGLE_TYPE = argument(
        '--camera',
        group=GROUP_SETTINGS,
        default='three_quarters',
        help='camera angle'
    )

    shader_style: SHADER_STYLE_TYPE = argument(
        '--style',
        default='plastic',
        group=GROUP_SETTINGS,
        help='Shader style to use'
    )

    title: str | None = argument(
        '--title',
        metavar='TEXT',
        default=None,
        group=GROUP_SETTINGS,
        help='title added to the top of the window'
    )

    source: str = argument(
        '-S', '--source',
        metavar='NAME',
        default='allen_mouse_10um',
        group=GROUP_SETTINGS,
        help='atlas source name. allen_human_500um as human'
    )

    root_alpha: float = argument(
        '--root-alpha',
        default=0.35,
        group=GROUP_SETTINGS,
        help='root alpha'
    )

    no_root: bool = argument(
        '--no-root',
        group=GROUP_SETTINGS,
        help='render without root(brain) mesh'
    )

    background: Literal['white', 'black'] = argument(
        '--bg',
        default='white',
        group=GROUP_SETTINGS,
        help='background color'
    )

    coordinate_space: Literal['ccf', 'brainrender'] = argument(
        '--coord-space',
        default='ccf',
        group=GROUP_SETTINGS,
        help='which coordinate space, by default ccf'
    )

    # ============= #
    # OPTIONAL VIEW #
    # ============= #

    GROUP_OPTIONAL = 'Optional Option'

    annotation: tuple[str, ...] | None = argument(
        '--annotation',
        validator.tuple().on_item(None, validator.str.match(r'^[^;]+:[^;]+:[^;]+$')) | validator.optional(),
        metavar='AV,DV,ML',
        type=str_tuple_type,
        default=None,
        group=GROUP_OPTIONAL,
        help='whether draw point annotation. e.g., 1.5:1:0.4,-3.2:0.8:0.4 for two points'
    )

    # ============ #
    # GROUP_REGION #
    # ============ #

    GROUP_REGION = 'Region Option'

    regions: str | tuple[str, ...] = argument(
        '-R', '--region',
        metavar='NAME,...',
        type=str_tuple_type,
        default=(),
        group=GROUP_REGION,
        help='region(s) name'
    )

    region_colors: str | tuple[str, ...] | None = argument(
        '--region-color',
        metavar='COLOR,...',
        type=str_tuple_type,
        default=None,
        group=GROUP_REGION,
        help='region(s) color'
    )

    regions_alpha: float = argument(
        '--region-alpha',
        validator.float.in_range_closed(0, 1),
        default=0.35,
        group=GROUP_REGION,
        help='region alpha value'
    )

    hemisphere: Literal['right', 'left', 'both'] = argument(
        '-H', '--hemisphere',
        default='both',
        group=GROUP_REGION,
        help='which hemisphere for rendering the region'
    )

    print_tree: bool = argument(
        '--print-tree',
        group=GROUP_REGION,
        help='print tree for the available regions for the given source'
    )

    tree_init: str | None = argument(
        '--tree-init',
        default=None,
        group=GROUP_REGION,
        help='init region for the tree print'
    )

    print_name: bool = argument(
        '--print-name',
        group=GROUP_REGION,
        help='print acronym and the corresponding name'
    )

    # ============ #
    # GROUP_OUTPUT #
    # ============ #

    GROUP_OUTPUT = 'Output Option'

    video_output: Path | None = argument(
        '--video-output',
        validator.path.is_suffix(['.mp4', '.avi']).optional(),
        default=None,
        group=GROUP_OUTPUT,
        help='video output path'
    )

    output: Path | None = argument(
        '-O', '--output',
        validator.path.is_suffix('.html').optional(),
        default=None,
        group=GROUP_OUTPUT,
        help='output path for the html, if None, preview'
    )

    #
    scene: brainrender.Scene
    logger = setup_clogger()
    _stop_render: bool = False  # flag for print mode

    def post_parsing(self):
        self._render_settings()
        self._verbose()

    def _render_settings(self):
        from brainrender import settings
        settings.BACKGROUND_COLOR = self.background
        settings.DEFAULT_ATLAS = self.source
        settings.ROOT_ALPHA = self.root_alpha
        settings.SHOW_AXES = False
        settings.WHOLE_SCREEN = False
        settings.DEFAULT_CAMERA = self.camera_angle
        settings.SHADER_STYLE = self.shader_style

        settings.vsettings.screenshot_transparent_background = True
        settings.vsettings.use_fxaa = False

    def _verbose(self):
        if self.print_tree:
            from neuralib.atlas.plot import plot_structure_tree
            plot_structure_tree(self.tree_init)
            self._stop_render = True

        if self.print_name:
            from neuralib.util.table import rich_data_frame_table
            file = self.get_atlas_brain_globe().root_dir / 'structures.csv'
            df = pl.read_csv(file).select('acronym', 'name')
            rich_data_frame_table(df)
            self._stop_render = True

    def run(self):
        self.post_parsing()

        if not self._stop_render:
            self.render()
            self.render_output()

    def render(self):
        """brainrender interactive"""
        self.scene = brainrender.Scene(root=not self.no_root, inset=False, title=self.title, screenshots_folder='.')
        self.scene.plotter.camera.Zoom(0.3)

        if self.annotation is not None:
            self._reconstruct_annotation()

        self._reconstruct_region()

    def render_output(self):
        """io handling. i.e., video, html output"""
        if self.video_output is not None:
            self.source = 'allen_mouse_25um'  # force set for whole brain scene
            self.video_maker(self.video_output)
        elif self.output is not None:
            self.export(self)
        else:
            self.scene.render()

    def _reconstruct_annotation(self):
        for ann in self.annotation:
            ap, dv, ml = tuple(map(float, ann.split(':')))
            dat = allen_to_brainrender_coord(np.array([ap, dv, ml]))  # (N, 3)
            self.scene.add(Points(dat, radius=120))

    def _reconstruct_region(self):
        color_list = self.region_colors or self.DEFAULT_REGION_COLORS

        if len(self.regions) != 0:
            for i, region in enumerate(self.regions):
                try:
                    color = color_list[i]
                except IndexError:
                    color = get_color(i, [''])

                self.logger.info(f'Plot Rois File: {i}, {region}, {color}')
                self.scene.add_brain_region(region, color=color, alpha=self.regions_alpha, hemisphere=self.hemisphere)

    @classmethod
    def export(cls, reconstructor: Self | None,
               output: Path | None = None,
               areas: list[str] | None = None,
               alpha: float = 0.15):
        """
        Export reconstruction as html

        :param reconstructor: `BrainRenderReconstructor` if use the current scene, and --output cli.
                    Otherwise, general func usage
        :param output: output file path
        :param areas: list of area(s)
        :param alpha: brain region alpha
        """

        if isinstance(reconstructor, BrainRenderCLI):
            scene = reconstructor.scene
            output = reconstructor.output
        else:
            scene = brainrender.Scene(inset=False, title='', screenshots_folder='.')
            output = output

        if areas is not None:
            if not isinstance(areas, list):
                raise TypeError('')

            for it in areas:
                scene.add_brain_region(it, alpha=alpha)

        scene.export(output)

    def video_maker(self, output_file: Path):
        """
        generate video

        :param output_file: video output path
        """
        from brainrender import VideoMaker

        d, f = output_file.parent, output_file.stem
        vm = VideoMaker(self.scene, save_fld=d, name=f)
        vm.make_video(azimuth=1, elevation=0, roll=0)

    def get_atlas_brain_globe(self, check_latest=False) -> BrainGlobeAtlas:
        return BrainGlobeAtlas(
            self.source,
            check_latest=check_latest
        )


if __name__ == '__main__':
    BrainRenderCLI().main()



from io import BufferedIOBase, BufferedReader
from pathlib import Path
from typing import TypeVar, Union, BinaryIO, Sequence

import numpy as np
import pandas as pd
import polars as pl
from matplotlib.axes import Axes
from numpy.typing import NDArray

__all__ = [
    'ArrayLike',
    'ArrayLikeStr',
    'AxesArray',
    #
    'PathLike',
    #
    'Series',
    'DataFrame'
]

# ===== #
# Array #
# ===== #

T = TypeVar('T')
"""Numeric"""

ArrayLike = Union[NDArray[T], Sequence[T], pd.Series, pl.Series]
"""Alias for array-like objects, including numpy arrays, lists, tuples, and series"""

ArrayLikeStr = Union[NDArray[np.str_], list[str], tuple[str, ...], pd.Series, pl.Series]
"""Alias for array-like objects of strings, including numpy arrays, lists, tuples, and series"""

AxesArray = Union[np.ndarray, list[Axes]]
"""Alias for matplotlib Axes numpy array"""

# ==== #
# Path #
# ==== #

PathLike = Union[str, Path, bytes, BinaryIO, BufferedIOBase, BufferedReader]
"""Alias for path-like objects, including strings, Paths, and file-like objects"""

# ================== #
# Series / DataFrame #
# ================== #

Series = Union[pd.Series, pl.Series]
"""Alias for series objects from pandas or polars"""

DataFrame = Union[pd.DataFrame, pl.DataFrame]
"""Alias for dataframe objects from pandas or polars"""



from typing import Literal

__all__ = ['CONFLICT_POLICY', 'FOREIGN_POLICY', 'UPDATE_POLICY']

CONFLICT_POLICY = Literal['rollback', 'abort', 'fail', 'ignore', 'replace']
""""""
FOREIGN_POLICY = Literal['SET NULL', 'SET DEFAULT', 'CASCADE', 'RESTRICT', 'NO ACTION']
""""""
UPDATE_POLICY = Literal['ABORT', 'FAIL', 'IGNORE', 'REPLACE', 'ROLLBACK']
""""""



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import Any, final, Generator, Literal

import numpy as np
import tifffile

from neuralib.typing import PathLike
from neuralib.util.verbose import fprint
from .core import AbstractScanner, DimCode

__all__ = ['lsm_file', 'TiffScanner']


@contextmanager
def lsm_file(filepath: PathLike) -> Generator[TiffScanner, None, None]:
    """context manager for load a lsm file

    :param filepath: lsm file path
    :return: :class:`~TiffScanner`
    """
    scanner = TiffScanner(filepath)
    try:
        yield scanner
    finally:
        scanner.close()


@final
class TiffScanner(AbstractScanner):
    """LSM confocal image data.

    **TODO**: multi-scene and different dimension data are lacking of testing
    """

    _image: np.ndarray | None
    _file_type: str

    def __init__(self, filepath: PathLike):
        self._image = tifffile.imread(filepath)
        self._file_type = Path(filepath).suffix

        super().__init__(filepath)

    def close(self):
        if self._image is not None:
            self._image = None

    def _load_metadata(self) -> dict[str, Any]:
        return parse_tif_meta(self._filepath, is_lsm=self.file_type == '.lsm')

    @property
    def image(self) -> np.ndarray:
        """image array"""
        return self._image

    @property
    def file_type(self) -> str:
        """file type of the image"""
        return self._file_type

    @property
    def dimcode(self) -> DimCode:
        match self.file_type:
            case '.lsm':
                if self.n_scenes == 1:
                    return 'ZCYX'
                else:
                    raise NotImplementedError('multi-scene not support yet')
            case _:
                raise NotImplementedError(f'{self.file_type} not support yet')

    @property
    def n_scenes(self) -> int:
        return self.metadata.get('DimensionP', 1)

    def get_channel_names(self, scene_idx=None) -> list[str]:
        return self.metadata.get('ChannelColors', {}).get('ColorNames', [])

    def view(self, channel: int = 0,
             depth: int | slice | np.ndarray | None = None,
             project_type: Literal['avg', 'max', 'min', 'std', 'median'] = 'max',
             norm: bool = True) -> np.ndarray:
        """
        Generates a view of the image data based on the provided parameters.
        Only provide a single scene view for now.

        :param channel: The channel index to select from the image. Defaults to 0.
        :param depth: Depth levels to process, which can be an integer, a slice,
            or a NumPy array. If None, all depth levels are used.
        :param project_type: Type of Z-projection to apply if multiple Z-slices are selected.
             Options include 'avg', 'max', 'min', 'std', 'median'. Defaults to 'max'.
        :param norm: A flag indicating whether to normalize the projected image by its
            maximum intensity value. Defaults to True.
        :return: A NumPy array representing the resulting image after applying depth
            projection, channel selection, and optional normalization.
        """

        if depth is None:
            depth = np.arange(self.image.shape[0])

        img = self.image[depth, channel, :, :]

        if isinstance(depth, int):
            pass
        elif isinstance(depth, (np.ndarray, slice)):
            img = self.z_projection(img, project_type)
        else:
            raise TypeError(f'unknown type {type(depth)}')

        if norm:
            img = img / np.max(img)

        return img


def parse_tif_meta(file: Path, **kwargs) -> dict[str, Any]:
    """
    Extracts metadata dict from a .tif/.tiff/.lsm file.
    Raises ValueError if the extension isn’t supported.

    :param file: Path to the file to be parsed.
    """
    suffix = file.suffix.lower()
    if suffix not in {'.tif', '.tiff', '.lsm'}:
        raise ValueError(f"Unsupported extension: {file.suffix!r}")

    meta_collect = {}
    with tifffile.TiffFile(file, **kwargs) as tif:
        # LSM metadata
        lsm = getattr(tif, 'lsm_metadata', None)
        if isinstance(lsm, dict):
            meta_collect.update(lsm)

        # ImageJ metadata
        ij = getattr(tif, 'imagej_metadata', None)
        if isinstance(ij, dict):
            meta_collect.update(ij)

        # Any other tuple of dicts you care about
        for attr in ('ome_metadata', 'ome_xml',):
            md = getattr(tif, attr, None)
            if isinstance(md, tuple):
                for part in md:
                    if not isinstance(part, dict):
                        fprint("Skipping non‐dict in %s: %r", attr, part)
                        continue
                    dup = meta_collect.keys() & part.keys()
                    if dup:
                        raise RuntimeError(f"Duplicate keys in {attr}: {dup}")
                    meta_collect.update(part)

    return meta_collect



from __future__ import annotations

from pathlib import Path

import numpy as np
from numpy.lib.npyio import NpzFile
from typing_extensions import Self, Literal

from neuralib.typing import PathLike

__all__ = [
    'STARDIST_MODEL',
    'read_stardist',
    'stardist_point_roi_helper',
    'StarDistSegmentation'
]

STARDIST_MODEL = Literal['2D_versatile_fluo', '2D_versatile_he', '2D_paper_dsb2018', '2D_demo']
"""stardist supported model type"""


def read_stardist(file: PathLike) -> StarDistSegmentation:
    """
    Read a cellpose segmentation result file

    :param file: stardist segmentation result ``.npy`` file
    :return: :class:`StarDistSegmentation`
    """
    return StarDistSegmentation.load(file)


def stardist_point_roi_helper(file: PathLike, output: PathLike) -> None:
    """Read a stardist segmentation result and convert the segmentation result to point coordinates

    :param file: stardist segmentation result ``.npz`` file
    :param output: ``*.roi`` output file path
    """
    StarDistSegmentation.load(file).to_roi(output)


class StarDistSegmentation:
    """`StarDist <https://github.com/stardist/stardist>`_ segmentation results

    `Dimension parameters`:

        N = Number of segmented cell

        E = Number of polygons edge

        W = Image width

        H = Image height

        P = Number of image pixel with label

    """

    def __init__(self, labels: np.ndarray,
                 cords: np.ndarray,
                 prob: np.ndarray,
                 filename: str,
                 model: STARDIST_MODEL):
        """

        :param labels: Image with label. `Array[float, [H, W]]`
        :param cords: Coordinates. `Array[float, [N, 2, E]]`
        :param prob: Detected probability. `Array[float, N]`
        :param filename: filepath of image
        :param model: :attr:`STARDIST_MODEL`
        """
        self._labels = labels
        self._cords = cords
        self._prob = prob

        self._filename = filename
        self._model = model

    @classmethod
    def load(cls, file: PathLike) -> Self:
        """
        Load a stardist segmentation result

        :param file: stardist segmentation result ``.npz`` file
        :return: :class:`StarDistSegmentation`
        """
        dat = np.load(file, allow_pickle=True)
        return cls(labels=cls._reconstruct_labels_from_index_value(dat),
                   cords=dat['cords'],
                   prob=dat['prob'],
                   filename=dat['filename'],
                   model=dat['model'])

    @classmethod
    def _reconstruct_labels_from_index_value(cls, dat: NpzFile) -> np.ndarray:
        h, w = dat['shape']
        index = dat['index']
        value = dat['value']

        image = np.full((h, w), np.nan)
        for i, (hi, wi) in enumerate(index):
            image[hi, wi] = value[i]

        return image

    @property
    def n_segmentation(self) -> int:
        """number of segmented cells"""
        return len(self._prob)

    @property
    def width(self) -> int:
        """image width"""
        return self._labels.shape[1]

    @property
    def height(self) -> int:
        """image height"""
        return self._labels.shape[0]

    @property
    def filename(self) -> Path:
        """filepath of image"""
        return Path(self._filename)

    @property
    def labels(self) -> np.ndarray:
        """Image with label. `Array[float, [H, W]]`"""
        return self._labels

    @property
    def cords(self) -> np.ndarray:
        """Coordinates. `Array[float, [N, 2, E]]`"""
        return self._cords

    @property
    def prob(self) -> np.ndarray:
        """Detected probability. `Array[float, N]`"""
        return self._prob

    @property
    def points(self) -> np.ndarray:
        """Coordinates to points by simple XY average. `Array[float, [N, 2]]`"""
        return self.cords.mean(axis=2)

    @property
    def model(self) -> STARDIST_MODEL:
        """stardist model type"""
        return self._model

    def mask_probability(self, threshold: float):
        """masking probability for the results

        :param threshold: probability threshold
        """
        mx = self.prob >= threshold
        self._cords = self._cords[mx]
        self._prob = self._prob[mx]

    # noinspection PyTypeChecker
    def to_npz(self, output_file: PathLike) -> None:
        """
        Save ``filename``, ``cord``, ``prob``, ``point``, ``shape``, ``index``, ``index``, ``value`` as a npz file.

        shape: `Array[int, 2] in H,W`

        index: index with labels. `Array[int, [P, 2]]`

        value: label value in its index `Array[int, P]`

        :param output_file: output ``*.npz`` file path
        """
        shape = np.array(self.labels.shape)
        index, value = self._get_index_value()

        np.savez(output_file,
                 cords=self.cords,
                 prob=self.prob,
                 points=self.points,
                 shape=shape,
                 index=index,
                 value=value,
                 filename=str(self.filename),
                 model=self.model)

    def _get_index_value(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Get labeled pixel index and its value.

        :return: `Array[int, [P, 2]]` and `Array[float, P]`
        """
        labels = self.labels
        mask = ~np.isnan(labels)
        index = np.argwhere(mask)
        value = labels[mask]

        return np.array(index).astype(np.int_), np.array(value).astype(np.int_)

    # noinspection PyTypeChecker
    def to_roi(self, output_file: PathLike) -> None:
        """Covert segmented roi to point roi, and save it as ``.roi`` for imageJ.

        :param output_file: ``*.roi`` output file path"""
        from roifile import ImagejRoi, ROI_TYPE, ROI_OPTIONS

        points = np.fliplr(self.points)  # XY rotate in .roi format
        roi = ImagejRoi(
            roitype=ROI_TYPE.POINT,
            options=ROI_OPTIONS.PROMPT_BEFORE_DELETING | ROI_OPTIONS.SUB_PIXEL_RESOLUTION,
            n_coordinates=self.points.shape[0],
            integer_coordinates=points,
            subpixel_coordinates=points
        )

        roi.tofile(output_file)



import random

from bokeh.model import Model
from bokeh.models import ColumnDataSource, GlyphRenderer
from bokeh.plotting import figure

from neuralib.dashboard import ViewComponent, View, BokehServer


class Graph(ViewComponent):
    data: ColumnDataSource
    render_dots: GlyphRenderer
    render_line: GlyphRenderer

    def __init__(self):
        self.x = []
        self.y = []
        self.data = ColumnDataSource(data=dict(x=[], y=[]))
        self.w = 0
        self.h = 0

    def plot(self, fig: figure, **kwargs):
        self.w = fig.width
        self.h = fig.height
        self.render_dots = fig.dot(
            x='x', y='y', source=self.data,
            size=self.w / 20)

        self.render_line = fig.line(
            x='x', y='y', source=self.data)

    def update(self):
        x = random.random() * self.w
        y = random.random() * self.h
        self.x.append(x)
        self.y.append(y)
        self.data.data = dict(x=self.x, y=self.y)


class Top(View):
    graph: Graph

    @property
    def title(self) -> str:
        return 'A'

    def setup(self) -> Model:
        fig = figure(
            width=800, height=800,
            x_range=(-10, 810),
            y_range=(-10, 810),
            toolbar_location='above')
        self.graph = Graph()
        self.graph.plot(fig)

        from bokeh.layouts import column
        return column(fig)

    def update(self):
        self.graph.update()
        self.run_timeout(1000, self.update)


if __name__ == '__main__':
    VIEW = Top()
    BokehServer().start(VIEW)



from __future__ import annotations

from typing import Sequence, Literal

import attrs
import numpy as np
from matplotlib.patches import Polygon
from typing_extensions import Self

from neuralib.util.unstable import unstable

__all__ = ['get_field_of_view',
           'get_cellular_coordinate',
           'FieldOfView',
           'CellularCoordinates']

UNIT = Literal['mm', 'um']


def get_field_of_view(am: Sequence[float],
                      pm: Sequence[float],
                      pl: Sequence[float],
                      al: Sequence[float],
                      *,
                      rotation_ml: float = 0,
                      rotation_ap: float = 0,
                      unit: UNIT = 'mm',
                      perpendicular: bool = True,
                      region_name: str | None = None) -> FieldOfView:
    """
    Construct a :class:`~FieldOfView` from four corner coordinates.

    :param am: anteromedial corner coordinate [x, y] in mm or µm.
    :param pm: posteromedial corner coordinate [x, y] in mm or µm.
    :param pl: posterolateral corner coordinate [x, y] in mm or µm.
    :param al: anterolateral corner coordinate [x, y] in mm or µm.
    :param rotation_ml: in-plane rotation (CCW) around the ML axis, in degrees.
    :param rotation_ap: tilt around the AP axis (foreshortening), in degrees.
    :param unit: unit of the corners value
    :param perpendicular: imaging objective is perpendicular to the cranial windows. if True, skip rotation and tilt transforms.
    :param region_name: optional identifier for this FOV region.
    :returns: FieldOfView instance with corners stacked and optionally transformed.
    """
    corners = np.vstack([am, pm, pl, al])
    return FieldOfView(corners, rotation_ml, rotation_ap, unit,
                       perpendicular=perpendicular,
                       region_name=region_name)


def _validator_corners(instance, attribute, value: np.ndarray):
    if not isinstance(value, np.ndarray):
        raise TypeError(f'{attribute} should be a numpy array')
    if value.shape != (4, 2):
        raise ValueError('')


@unstable()
def _rotate_and_tilt(pts: np.ndarray, ml: float, ap: float) -> np.ndarray:
    """telecentric / orthographic imaging transformation"""
    pivot = pts.mean(axis=0)
    theta_x = np.deg2rad(ml)
    c, s = np.cos(theta_x), np.sin(theta_x)
    R = np.array([[c, -s], [s, c]])
    pts0 = (pts - pivot) @ R.T + pivot
    theta_y = np.deg2rad(ap)
    scale_x = np.cos(theta_y)
    pts1 = (pts0 - pivot) * np.array([scale_x, 1.0]) + pivot

    return pts1


@attrs.frozen
class FieldOfView:
    """Field‑Of‑View defined by its four XY corners"""

    corners: np.ndarray = attrs.field(validator=_validator_corners)
    """corner coordinates in order [AM, PM, PL, AL]. `Array[float, [4, 2]]`"""
    rotation_ml: float = attrs.field(default=0.0)
    """in-plane rotation (CCW) around the ML axis, in degrees"""
    rotation_ap: float = attrs.field(default=0.0)
    """tilt around the AP axis (foreshortening), in degrees"""
    unit: UNIT = attrs.field(default='mm', validator=attrs.validators.in_(['mm', 'um']))
    """unit of the corners value"""
    perpendicular: bool = attrs.field(default=True, kw_only=True)
    """imaging objective is perpendicular to the cranial windows. if True, skip rotation and tilt transforms"""
    region_name: str | None = attrs.field(default=None, kw_only=True)
    """optional identifier for this FOV region"""

    def __attrs_post_init__(self):
        if not self.perpendicular:
            new_corners = _rotate_and_tilt(
                self.corners,
                self.rotation_ml,
                self.rotation_ap,
            )
            object.__setattr__(self, 'corners', new_corners)

    @property
    def am(self) -> np.ndarray:
        """anteromedial corner coordinate [x, y] in mm or µm."""
        return self.corners[0]

    @property
    def pm(self) -> np.ndarray:
        """posteromedial corner coordinate [x, y] in mm or µm"""
        return self.corners[1]

    @property
    def pl(self) -> np.ndarray:
        """posterolateral corner coordinate [x, y] in mm or µm"""
        return self.corners[2]

    @property
    def al(self) -> np.ndarray:
        """anterolateral corner coordinate [x, y] in mm or µm"""
        return self.corners[3]

    @property
    def ap_distance(self) -> float:
        """span along the AP axis"""
        return float(np.ptp(self.corners[:, 1]))

    @property
    def ml_distance(self) -> float:
        """span along the ML axis"""
        return float(np.ptp(self.corners[:, 0]))

    def invert_axes(self, ap: bool = True, ml: bool = True) -> Self:
        """Return a new FOV with specified axes inverted

        :param ap: invert anterior-posterior (Y) axis if True.
        :param ml: invert medial-lateral (X) axis if True.
        """
        factors = np.array([-1. if ml else 1., -1. if ap else 1.])
        return attrs.evolve(self, corners=self.corners * factors)

    def to_um(self) -> Self:
        if self.unit == 'um':
            raise RuntimeError('unit already in um')
        return attrs.evolve(self, corners=self.corners * 1000, unit='um')

    def to_polygon(self, **kwargs) -> Polygon:
        """convert corners to a matplotlib Polygon patch"""
        idx = [0, 1, 3, 2]
        reorder = self.corners[idx]
        return Polygon(reorder, closed=True, edgecolor='r', facecolor='none', **kwargs)


def get_cellular_coordinate(neuron_idx: np.ndarray,
                            ap: np.ndarray,
                            ml: np.ndarray, *,
                            unit: UNIT = 'mm',
                            plane_index: np.ndarray | None = None) -> CellularCoordinates:
    """
    Get cellular coordinates container for doing brain mapping / topographical analysis

    :param neuron_idx: neuron index. `Array[float, N]`
    :param ap: anterior posterior coordinates. `Array[float, N]`
    :param ml: medial lateral coordinates. `Array[float, N]`
    :param unit: unit of the ap/ml value. default in `mm`
    :param plane_index: neuron's corresponding image plane. `Array[float, N]`. If None then full_zero
    :return: :class:`~CellularCoordinates`
    """
    return CellularCoordinates(neuron_idx, ap, ml, unit, plane_index)


def _validator_shape(instance: CellularCoordinates, attribute, value: np.ndarray | None):
    if value is None:
        return
    else:
        assert instance.neuron_idx.shape == value.shape


@attrs.define
class CellularCoordinates:
    """Cellular Coordinates container"""

    neuron_idx: np.ndarray
    """neuron index. `Array[float, N]`"""
    ap: np.ndarray
    """anterior posterior coordinates (default in mm). `Array[float, N]`"""
    ml: np.ndarray
    """medial lateral coordinates (default in mm). `Array[float, N]`"""
    unit: UNIT = attrs.field(default='mm', validator=attrs.validators.in_(['mm', 'um']))
    """unit of the ap/ml value"""
    plane_index: np.ndarray = attrs.field(default=None, validator=_validator_shape)
    """neuron's corresponding image plane. `Array[float, N]`"""
    value: np.ndarray | None = attrs.field(default=None, validator=_validator_shape)
    """metric (i.e., used in topographical analysis). `Array[float, N]`"""

    def __attrs_post_init__(self):
        if self.plane_index is None:
            self.plane_index = np.full_like(self.neuron_idx, 0, dtype=int)

        assert self.neuron_idx.shape == self.ap.shape == self.ml.shape == self.plane_index.shape

    def relative_origin(self, fov: FieldOfView,
                        origin: Literal['am', 'pm', 'al', 'pl'] = 'am') -> Self:
        """
        coordinates relative to :class:`~FieldOfView` origin point

        :param fov: :class:`~FieldOfView`
        :param origin: relative origin point
        :return:
        """

        factor = 1000 if self.unit == 'mm' else 1
        ap_um = self.ap * factor
        ml_um = self.ml * factor
        if fov.unit != 'um':
            fov = fov.to_um()

        orig_pt = getattr(fov, origin)
        pts = np.vstack([ml_um, ap_um]).T

        if fov.perpendicular:
            delta = orig_pt - pts  # posterior to origin
        else:
            delta_pts = orig_pt - _rotate_and_tilt(
                pts,
                ml=fov.rotation_ml,
                ap=fov.rotation_ap,
            )
            delta = delta_pts

        ml_new, ap_new = delta[:, 0], delta[:, 1]
        return attrs.evolve(self, ml=ml_new, ap=ap_new, unit='um')

    def with_value(self, value: np.ndarray) -> Self:
        """assign value foreach neuron"""
        return attrs.evolve(self, value=value)

    def with_masking(self, mask: np.ndarray) -> Self:
        """do neuronal selection by bool masking
        :param mask: `Array[bool, N]`
        """
        return attrs.evolve(
            self,
            neuron_idx=self.neuron_idx[mask],
            ap=self.ap[mask],
            ml=self.ml[mask],
            plane_index=self.plane_index[mask],
            value=None if self.value is None else self.value[mask]
        )



import pickle
from pathlib import Path
from typing import Literal

import numpy as np
from scipy.stats import zscore

from neuralib.typing import PathLike
from .core import RasterOptions, RasterMapResult

__all__ = ['DATA_TYPE', 'run_rastermap']

DATA_TYPE = Literal['cellular', 'wfield']


def run_rastermap(neural_activity: np.ndarray,
                  bin_size: int,
                  dtype: DATA_TYPE, *,
                  options: RasterOptions | None = None,
                  svd_components: int = 128,
                  svd_cache: PathLike | None = None,
                  invalid_svd_cache: bool = False,
                  filename: str | None = None,
                  save_path: str | None = None,
                  **kwargs) -> RasterMapResult:
    """
    Run rastermap

    :param neural_activity: neural activity. `Array[float, [N, T]]` | `Array[Any, [T, H, W]]`
    :param bin_size: number of neurons to bin over
    :param dtype: :attr:`~neuralib.model.rastermap.run.DATA_TYPE`
    :param options: :class:`~neuralib.model.rastermap.core.RasterOptions`
    :param svd_components: number of SVD components to use (for widefield dtype)
    :param svd_cache: path to cache SVD file (for widefield dtype)
    :param invalid_svd_cache: invalid (recompute) SVD cached file (for widefield dtype)
    :param filename: optional specify if GUI re-launch
    :param save_path: optional specify if GUI re-launch
    """
    rastermap = RunRastermap(neural_activity, bin_size, dtype,
                             options=options,
                             svd_components=svd_components,
                             svd_cache=svd_cache,
                             invalid_svd_cache=invalid_svd_cache,
                             filename=filename,
                             save_path=save_path)
    return rastermap.run(**kwargs)


class RunRastermap:
    DEFAULT_OPTIONS: RasterOptions = {
        'n_clusters': 50,
        'n_PCs': 128,
        'locality': 0.75,
        'time_lag_window': 5,
        'grid_upsample': 10,
    }

    def __init__(self, neural_activity: np.ndarray,
                 bin_size: int,
                 dtype: DATA_TYPE, *,
                 options: RasterOptions | None = None,
                 svd_components: int = 128,
                 svd_cache: PathLike | None = None,
                 invalid_svd_cache: bool = False,
                 filename: str | None = None,
                 save_path: str | None = None):
        """

        :param neural_activity: neural activity. `Array[float, [N, T]]` | `Array[Any, [T, H, W]]`
        :param bin_size: number of neurons to bin over
        :param options: :class:`~neuralib.model.rastermap.core.RasterOptions`
        :param svd_components: number of SVD components to use (for widefield dtype)
        :param svd_cache: path to cache SVD file (for widefield dtype)
        :param invalid_svd_cache: invalid (recompute) SVD cached file (for widefield dtype)
        :param filename: optional specify if GUI re-launch
        :param save_path: optional specify if GUI re-launch
        """

        self.neural_activity = neural_activity
        self.bin_size = bin_size
        self.dtype = dtype

        self.svd_components = svd_components
        self.svd_cache = svd_cache
        self.invalid_svd_cache = invalid_svd_cache

        self.filename = filename
        self.save_path = save_path

        self._options = options or self.DEFAULT_OPTIONS

    @property
    def options(self) -> RasterOptions:
        return self._options

    def run(self, **kwargs) -> RasterMapResult:
        """run rastermap with the given dtype"""
        match self.dtype:
            case 'cellular':
                ret = self.run_cellular(**kwargs)
            case 'wfield':
                ret = self.run_wfield(**kwargs)
            case _:
                raise NotImplementedError('')

        if self.save_path is not None:
            ret.save(self.save_path)

        return ret

    def run_cellular(self, **kwargs) -> RasterMapResult:
        """cellular input run. `Array[float, [N, T]]`"""
        try:
            from rastermap import Rastermap
            import rastermap.utils
        except ImportError:
            raise RuntimeError('pip install rastermap first')

        opt = self.options
        model = Rastermap(n_clusters=opt['n_clusters'],
                          n_PCs=opt['n_PCs'],
                          locality=opt['locality'],
                          time_lag_window=opt['time_lag_window'],
                          grid_upsample=opt['grid_upsample'],
                          **kwargs)

        model.fit(self.neural_activity)

        embedding = model.embedding
        isort = model.isort

        sn = rastermap.utils.bin1d(self.neural_activity[isort], bin_size=self.bin_size, axis=0)

        ret = RasterMapResult(
            filename=self.filename,
            save_path=self.save_path,
            isort=isort,
            embedding=embedding,
            ops=opt,
            user_clusters=[],
            super_neurons=sn
        )

        return ret

    def run_wfield(self, **kwargs):
        """wfield input run. `Array[Any, [T, H, W]]`"""
        from neuralib.imaging.widefield import compute_singular_vector

        try:
            from rastermap import Rastermap
            import rastermap.utils
        except ImportError:
            raise RuntimeError('pip install rastermap first')

        #
        opt = self.options
        model = Rastermap(
            n_clusters=opt['n_clusters'],
            n_PCs=opt['n_PCs'],
            locality=opt['locality'],
            time_lag_window=opt['time_lag_window'],
            grid_upsample=opt['grid_upsample'],
            **kwargs
        )

        # svd cache
        if self.svd_cache is not None:
            file = Path(self.svd_cache)
            if file.exists() and not self.invalid_svd_cache:
                with file.open("rb") as f:
                    sv = pickle.load(f)
            else:
                sv = compute_singular_vector(self.neural_activity, self.svd_components)
                with file.open('wb') as f:
                    pickle.dump(sv, f)
        else:
            sv = compute_singular_vector(self.neural_activity, self.svd_components)

        usv = sv.singular_value * sv.left_vector
        vsv = sv.right_vector

        # fit
        model.fit(Usv=usv, Vsv=vsv)
        embedding = model.embedding
        isort = model.isort
        Vsv_sub = model.Vsv

        U_sn = rastermap.utils.bin1d(sv.left_vector[isort], bin_size=self.bin_size, axis=0)
        sn = U_sn @ Vsv_sub.T
        sn = zscore(sn, axis=1)

        ret = RasterMapResult(
            filename=self.filename,
            save_path=self.save_path,
            isort=isort,
            embedding=embedding,
            ops=opt,
            user_clusters=[],
            super_neurons=sn
        )

        return ret



import numpy as np

from neuralib.plot import plot_figure
from neuralib.plot.colormap import insert_colorbar, insert_cyclic_colorbar
from neuralib.typing import PathLike
from .fft import SequenceFFT

__all__ = ['plot_retinotopic_maps']


def plot_retinotopic_maps(sequence: np.ndarray, *,
                          output: PathLike | None = None,
                          interp: str = 'none',
                          intensity_cmap='binary',
                          phase_cmap='hsv',
                          **kwargs):
    """
    Plot retinotopic maps based on fft calculation.

    .. seealso::

        :class:`~neuralib.imaging.widefield.fft.SequenceFFT`

    :param sequence: Image sequence. `Array[float | uint8, [F, H, W]]`
    :param output: Output path for the figure, defaults is None for ``show()``
    :param interp: Kwarg interpolation for ``ax.imshow()``
    :param intensity_cmap: Intensity color map, defaults to 'binary'
    :param phase_cmap: Intensity phase color map, defaults to 'hsv'
    :param kwargs: Additional arguments passed to :meth:`~neuralib.imaging.widefield.fft.SequenceFFT.as_colormap()`
    """
    seq_fft = SequenceFFT(sequence)

    with plot_figure(output, 1, 3, figsize=(12, 6)) as _ax:
        ax = _ax[0]
        im = ax.imshow(seq_fft.get_intensity(), cmap=intensity_cmap, interpolation=interp)
        insert_colorbar(ax, im)
        ax.set_title('intensity')
        ax.axis('off')

        ax = _ax[1]
        im = ax.imshow(seq_fft.get_phase(), cmap=phase_cmap, interpolation=interp)
        insert_cyclic_colorbar(ax, im, num_colors=36, width=0.2, inner_diameter=1, vmin=0, vmax=1)
        ax.set_title('phase')
        ax.axis('off')

        ax = _ax[2]
        im = ax.imshow(seq_fft.as_colormap(**kwargs), cmap='hsv', interpolation=interp)
        insert_cyclic_colorbar(ax, im, num_colors=36, width=0.5, inner_diameter=1, vmin=0, vmax=1)
        ax.set_title('retinotopic map')
        ax.axis('off')



import abc
import datetime
from pathlib import Path
from typing import Generic, TypeVar

import numpy as np

from neuralib.typing import is_iterable
from neuralib.util.verbose import fprint
from .persistence import ensure_persistence_class

__all__ = [
    'create_date_validate',
    'attributes_validate',
    #
    'ETLConcatable',
    'validate_concat_etl_persistence',
]

T = TypeVar('T')


def _to_datetime(obj: float | Path | tuple[int, int, int] | datetime.date | datetime.datetime) -> datetime.datetime:
    if isinstance(obj, float):
        return datetime.datetime.fromtimestamp(obj)
    elif isinstance(obj, tuple):
        return datetime.datetime(obj[0], obj[1], obj[2])
    elif isinstance(obj, Path):
        if not obj.exists():
            raise FileNotFoundError
        return datetime.datetime.fromtimestamp(obj.stat().st_mtime)
    elif isinstance(obj, (datetime.datetime, datetime.date)):
        return obj
    else:
        raise TypeError(str(type(obj).__name__))


def create_date_validate(obj: float | Path | tuple[int, int, int] | datetime.date | datetime.datetime,
                         ref: float | Path | tuple[int, int, int] | datetime.date | datetime.datetime,
                         verbose=True) -> bool:
    """test *obj* is created after *ref*.

    :param obj: time stamp, file path or date instance
    :param ref: time stamp, file path or date instance
    :param verbose:
    :return: Is *obj* file newer than *ref* file?
    """
    try:
        obj_time = _to_datetime(obj)
    except FileNotFoundError:
        if verbose:
            fprint(f'obj file not existed : {obj}')
        return False

    try:
        ref_time = _to_datetime(ref)
    except FileNotFoundError:
        if verbose:
            fprint(f'ref file not existed : {ref}')
        return False

    if obj_time >= ref_time:
        return True

    if verbose:
        if isinstance(ref, float):
            msg = str(datetime.date.fromtimestamp(obj))
        elif isinstance(obj, tuple):
            msg = str(datetime.date(obj[0], obj[1], obj[2]))
        elif isinstance(ref, Path):
            msg = ref.name
        else:
            msg = str(ref)

        fprint(f'persistence data is created before {msg}')

    return False


def attributes_validate(obj: T, *exclude: str) -> bool:
    info = ensure_persistence_class(obj)
    for field in info.fields:
        attr = field.field_name
        if attr not in exclude and not hasattr(obj, attr) and not field.optional:
            return False
    return True


# ========================= #
# User-Specific for 2P data #
# ========================= #

class ETLConcatable(Generic[T], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def concat_etl(self, data: list[T]) -> T:
        pass


class PersistenceConcatError(Exception):
    pass


def validate_concat_etl_persistence(data: list[T], field_check: tuple[str, ...] | None = None) -> None:
    """

    :param data: list of persistence instance
    :param field_check: field name under persistence cls for checking
    :return:
    """
    #
    if not hasattr(data[0], 'exp_date') or not hasattr(data[0], 'animal'):
        raise RuntimeError('different dataset is not concatable')

    #
    for it in data:

        if not hasattr(it, 'plane_index'):
            raise AttributeError('not plane index field')

        if not isinstance(it.plane_index, int):
            raise TypeError('')

    if len(set([it.exp_date for it in data])) != 1:
        print(set([it.exp_date for it in data]))
        raise PersistenceConcatError('different exp date')

    if len(set([it.animal for it in data])) != 1:
        raise PersistenceConcatError('different animal')

    #
    if field_check is not None:
        for f in field_check:
            init = getattr(data[0], f)

            for it in data:
                check = getattr(it, f)

                if not is_iterable(init):
                    if check != init:
                        raise PersistenceConcatError(f' field:{f} not consistent')

                elif is_iterable(init) and isinstance(init, np.ndarray):
                    if not np.array_equal(init, check):
                        raise PersistenceConcatError(f' field:{f} not consistent')

                else:
                    for i, v in enumerate(init):
                        check_field = check
                        if check_field[i] != v:
                            raise PersistenceConcatError(f' field:{f} not consistent')



from pathlib import Path

from neuralib.atlas.data import load_bg_structure_tree
from neuralib.typing import PathLike

__all__ = ['plot_sunburst_acronym',
           'plot_structure_tree']


def plot_sunburst_acronym(output: PathLike | None = None):
    """
    Plot allen brain structure tree interactive plot

    :param output: figure output path, otherwise, render interactively
    :return:
    """
    from plotly import express as px

    data = load_bg_structure_tree(paired=True)
    data = dict(names=data['names'], parents=data['parents'])

    #
    fig = px.sunburst(
        data,
        names='names',
        parents='parents',

    )
    if output is not None:
        fig.write_image(output)
    else:
        fig.show()


def plot_structure_tree(starting_node: str | None = None, output: PathLike | None = None) -> None:
    """Show tree for the brain structure

    :param starting_node: starting node for rendering the tree. If None, then plot all
    :param output: output file txt. print if None
    """
    from anytree import Node, RenderTree

    df = load_bg_structure_tree(paired=True)

    nodes = {}
    for it in df.iter_rows(named=True):
        name = it['acronym']
        parent = it['parent_acronym']

        if name not in nodes:
            nodes[name] = Node(name)
        if parent not in nodes:
            nodes[parent] = Node(parent)

        nodes[name].parent = nodes[parent]

    #
    start = starting_node if starting_node is not None else 'root'
    if start not in nodes:
        raise ValueError(f'{starting_node} not find in the structure tree')

    #
    for pre, fill, node in RenderTree(nodes[start]):
        if output is not None:
            output = Path(output)

            if not output.exists():
                with Path(output).open('a') as file:
                    print("%s%s" % (pre, node.name), file=file)
            else:
                raise FileExistsError(f'{output}')
        else:
            print("%s%s" % (pre, node.name))



from __future__ import annotations

from typing import Literal, Sequence

import cv2
import numpy as np
from typing_extensions import Self

from neuralib.typing import PathLike

__all__ = ['image_array',
           'ImageArrayWrapper']

RGB_CHANNEL_TYPE = Literal['r', 'g', 'b', 'red', 'green', 'blue']


def image_array(dat: np.ndarray | PathLike, *,
                mode: Literal['RGB', 'RGBA', 'gray'] | None = None,
                alpha: bool = False) -> ImageArrayWrapper:
    """
    Image array numpy subclass

    :param dat: Image data as a NumPy array or a file path
    :param mode: Color mode {'RGB', 'RGBA', 'gray'}. Optional if dat is ndarray.
    :param alpha: If True and loading from file, convert to RGBA
    :return:
    """
    return ImageArrayWrapper(dat, mode=mode, alpha=alpha)


class ImageArrayWrapper(np.ndarray):
    """Subclass of numpy.ndarray that wraps an image and provides chainable image processing methods"""

    def __new__(cls, dat: np.ndarray | PathLike, *,
                mode: Literal['RGB', 'RGBA', 'gray'] | None = None,
                alpha: bool = False) -> Self:
        """

        :param dat: Image data as a NumPy array or a file path
        :param mode: Color mode {'RGB', 'RGBA', 'gray'}. Optional if dat is ndarray.
        :param alpha: If True and loading from file, convert to RGBA
        """
        if isinstance(dat, PathLike):
            img = cv2.imread(str(dat))
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            if alpha:
                img = cv2.cvtColor(img, cv2.COLOR_RGB2RGBA)
                mode = 'RGBA'
            else:
                mode = 'RGB'
            dat = img
        else:
            dat = np.asarray(dat)

            match (mode, dat.ndim, dat.shape[-1] if dat.ndim == 3 else None):
                case (None, 2, _):
                    mode = 'gray'
                case (None, 3, 3):
                    mode = 'RGB'
                case (None, 3, 4):
                    mode = 'RGBA'
                case ('gray', 2, _):
                    pass  # valid grayscale
                case ('RGB', 3, 3):
                    pass  # valid rgb
                case ('RGBA', 3, 4):
                    pass  # valid rgba
                case _:
                    raise ValueError(f"Invalid shape {dat.shape} for mode={mode!r}")

        obj = dat.view(cls)
        obj.mode = mode
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.mode = getattr(obj, 'mode', 'RGB')

    @property
    def height(self) -> int:
        """image height"""
        return self.shape[0]

    @property
    def width(self) -> int:
        """image width"""
        return self.shape[1]

    def to_gray(self) -> Self:
        """convert the image array to grayscale"""
        if self.ndim == 2:
            ret = self.copy().astype("uint8")
        elif self.shape[2] == 3:
            ret = cv2.cvtColor(self, cv2.COLOR_RGB2GRAY)
        elif self.shape[2] == 4:
            ret = cv2.cvtColor(self, cv2.COLOR_RGBA2GRAY)
        else:
            raise RuntimeError(f'Unexpected image shape: {self.shape} for grayscale conversion')

        return ImageArrayWrapper(ret, mode='gray')

    def flipud(self) -> Self:
        """flip the image array upside down (vertical)"""
        return ImageArrayWrapper(np.flipud(self))

    def fliplr(self) -> Self:
        """flip the image array left to right (horizontal flip)"""
        return ImageArrayWrapper(np.fliplr(self))

    def select_channel(self, channel: RGB_CHANNEL_TYPE) -> Self:
        """extract a single color channel from an RGB or RGBA image

        :param channel: one of 'r'/'red', 'g'/'green', or 'b'/'blue'.
        """
        if self.ndim < 3:
            raise RuntimeError(f'Image shape invalid for splitting channel: {self.ndim}')

        channels = cv2.split(self)
        match channel:
            case 'r' | 'red':
                ret = channels[0]
            case 'g' | 'green':
                ret = channels[1]
            case 'b' | 'blue':
                ret = channels[2]
            case _:
                raise ValueError(f'invalid {channel} argument')

        return ImageArrayWrapper(ret, mode='gray')

    def view_2d(self, flipud: bool = False) -> Self:
        """
        Convert a multi-channel image to a 2D representation.

        - For a 4-channel image, the array is reinterpreted as a 2D array of 32-bit integers.
        - For a 3-channel image, the result is a grayscale image obtained by applying luminance conversion.

        :param flipud: reverse the order of elements along axis 0 (up/down)
        """
        if self.ndim == 2:
            return self

        match self.shape[2]:
            case 4:
                w, h, _ = self.shape
                ret = self.view(dtype=np.uint32).reshape((w, h))
            case 3:
                r, g, b = self[:, :, 0], self[:, :, 1], self[:, :, 2]
                ret = (0.2989 * r + 0.5870 * g + 0.1140 * b).astype(np.uint8)
            case _:
                raise ValueError(f'invalid arr shape: {self.shape}')

        zelf = ImageArrayWrapper(ret, mode='gray')

        if flipud:
            zelf = zelf.flipud()

        return zelf

    def gaussian_blur(self, ksize: Sequence[int], sigma_x: float, sigma_y: float, **kwargs) -> Self:
        """
        Apply a Gaussian blur to the image.

        :param ksize: Kernel size (e.g., (5, 5)). The width and height should be odd numbers.
        :param sigma_x: Standard deviation in the X direction.
        :param sigma_y: Standard deviation in the Y direction.
        :param kwargs: Additional keyword arguments for ``cv2.GaussianBlur()``.
        """
        img = cv2.GaussianBlur(self, ksize=ksize, sigmaX=sigma_x, sigmaY=sigma_y, **kwargs)
        return ImageArrayWrapper(img)

    def canny_filter(self, threshold_1: float = 30, threshold_2: float = 150, **kwargs) -> Self:
        """
        Apply the Canny edge detection algorithm to the grayscale version of the image.

        :param threshold_1: The first threshold for the hysteresis procedure.
        :param threshold_2: The second threshold for the hysteresis procedure.
        :param kwargs: Additional keyword arguments for ``cv2.Canny()``.
        """
        img = cv2.Canny(self.to_gray(), threshold1=threshold_1, threshold2=threshold_2, **kwargs)
        return ImageArrayWrapper(img)

    def binarize(self, thresh: float, maxval: float = 255, **kwargs) -> Self:
        """
        Convert the image to a binary image using a fixed threshold.

        :param thresh: Threshold value. Pixels above this value are set to maxval; otherwise, 0.
        :param maxval: The value to use for pixels above the threshold.
        :param kwargs: Additional keyword arguments for ``cv2.threshold()``.
        """
        _, img = cv2.threshold(self, thresh, maxval=maxval, type=cv2.THRESH_BINARY, **kwargs)
        return ImageArrayWrapper(img)

    def denoise(self, h: int = 10, temp_win_size: int = 7, search_win_size: int = 21, **kwargs) -> Self:
        """
        Apply Non-local Means Denoising to the image.

        - For grayscale images, cv2.fastNlMeansDenoising is used.
        - For color images, cv2.fastNlMeansDenoisingColored is used.

        :param h: Filtering parameter controlling the degree of smoothing.
        :param temp_win_size: Template window size in pixels.
        :param search_win_size: Search window size in pixels.
        :param kwargs: Additional keyword arguments for the cv2 denoising function.
        """
        if self.mode == 'gray':
            fn = getattr(cv2, 'fastNlMeansDenoising')
        else:
            fn = getattr(cv2, 'fastNlMeansDenoisingColored')
        img = fn(self, h=h, templateWindowSize=temp_win_size, searchWindowSize=search_win_size, **kwargs)
        return ImageArrayWrapper(img)

    def enhance_contrast(self) -> Self:
        """Enhance the contrast of the image using histogram equalization"""
        match self.mode:
            case 'gray':
                eq = cv2.equalizeHist(self)
                return ImageArrayWrapper(eq, mode='gray')
            case 'RGB':
                ycrcb = cv2.cvtColor(self, cv2.COLOR_RGB2YCrCb)
                ycrcb[:, :, 0] = cv2.equalizeHist(ycrcb[:, :, 0])
                eq = cv2.cvtColor(ycrcb, cv2.COLOR_YCrCb2RGB)
                return ImageArrayWrapper(eq, mode='RGB')
            case 'RGBA':
                rgb = self[..., :3]
                alpha = self[..., 3]
                ycrcb = cv2.cvtColor(rgb, cv2.COLOR_RGB2YCrCb)
                ycrcb[:, :, 0] = cv2.equalizeHist(ycrcb[:, :, 0])
                eq_rgb = cv2.cvtColor(ycrcb, cv2.COLOR_YCrCb2RGB)
                eq = np.dstack((eq_rgb, alpha))
                return ImageArrayWrapper(eq, mode='RGBA')
            case _:
                raise ValueError(f'invalid mode: {self.mode}')

    def local_maxima(self, channel: RGB_CHANNEL_TYPE, **kwargs) -> Self:
        """
        Compute the local maxima of the image on a specified color channel.

        The specified channel is first extracted (and returned as a grayscale image),
        then the skimage local_maxima function is applied.
        :param channel: one of 'r'/'red', 'g'/'green', or 'b'/'blue'.
        :param kwargs: additional keyword arguments for ``skimage.morphology.local_maxima()``.
        """
        from skimage.morphology import local_maxima

        img = self.select_channel(channel)

        if np.sum(img) == 0:
            return ImageArrayWrapper(np.zeros_like(img, dtype=np.uint8), mode='gray')
        else:
            return ImageArrayWrapper(local_maxima(img, **kwargs), mode='gray')



import numpy as np


def filter_outliers(x, y, filter_window=15, baseline_window=50, max_spike=25, max_diff=2):
    """from ``facemap.utils``"""
    # remove frames with large jumps
    x_diff = np.abs(np.append(np.zeros(1, ), np.diff(x), ))
    y_diff = np.abs(np.append(np.zeros(1, ), np.diff(y), ))
    replace_inds = np.logical_or(x_diff > max_diff, y_diff > max_diff)
    x[replace_inds] = np.nan
    y[replace_inds] = np.nan

    # remove frames with large deviations from baseline
    x_baseline = nanmedian_filter(x, baseline_window)
    y_baseline = nanmedian_filter(y, baseline_window)
    replace_inds = np.logical_or(
        np.abs(x - x_baseline) > max_spike, np.abs(y - y_baseline) > max_spike
    )
    x[replace_inds] = np.nan
    y[replace_inds] = np.nan
    replace_inds = np.isnan(x)

    # filter x and y
    x_filt = nanmedian_filter(x, filter_window)
    y_filt = nanmedian_filter(y, filter_window)

    # this in theory shouldn't add more frames
    replace_inds = np.logical_or(replace_inds, np.isnan(x_filt))
    ireplace = np.nonzero(replace_inds)[0]

    # replace outlier frames with median
    if len(ireplace) > 0:
        # good indices
        iinterp = np.nonzero(np.logical_and(~replace_inds, ~np.isnan(x_filt)))[0]
        x[replace_inds] = np.interp(ireplace, iinterp, x_filt[iinterp])
        y[replace_inds] = np.interp(ireplace, iinterp, y_filt[iinterp])

    return x, y


def nanmedian_filter(x, win=7):
    """nanmedian filter array along last axis"""
    nt = x.shape[-1]
    # pad so that x will be divisible by win
    pad = (win - (nt + 2 * (win // 2)) % win) % win
    xpad = np.pad(x, (win // 2, win // 2 + win + pad), mode="edge")
    xmed = np.zeros_like(x)
    for k in range(win):
        xm = np.nanmedian(xpad[k: k - win].reshape(-1, win), axis=-1)
        xmed[..., k::win] = xm[: len(np.arange(k, nt, win))]
    return xmed



from __future__ import annotations

import abc
from typing import Callable, Literal, overload, TypeVar, Generic, Any, TYPE_CHECKING

import numpy as np
import polars as pl
from polars.dataframe.group_by import GroupBy
from polars.testing import assert_frame_equal

from neuralib.util.verbose import printdf

if TYPE_CHECKING:
    from collections.abc import Sequence, Iterable, Collection
    from typing_extensions import Self, ParamSpec, Concatenate
    from polars import _typing as pty

    P = ParamSpec('P')

__all__ = ['DataFrameWrapper',
           'helper_with_index_column',
           'assert_polars_equal_verbose']


class DataFrameWrapper(metaclass=abc.ABCMeta):
    """
    Abstract wrapper class for a `polars.DataFrame`, enabling convenient and composable
    dataframe operations in a subclassable, object-oriented interface.

    This base class is intended to be inherited by custom data structures whose core data
    is represented as a `polars.DataFrame`. It provides a suite of standard dataframe
    operations (e.g., filtering, sorting, renaming, joining) that return the wrapper
    instance (`Self`), preserving method chaining and encapsulation.

    This allows users to write clean, expressive logic using their custom wrapper class
    while still leveraging the full power of Polars.

    Subclasses **must** implement the `dataframe` method to get or set the internal
    `polars.DataFrame`.

    Examples
    --------
    A minimal subclass that wraps a Polars DataFrame:

    >>> class MyTable(DataFrameWrapper):
    ...     def __init__(self, data: pl.DataFrame):
    ...         self._data = data
    ...
    ...     def dataframe(self, dataframe: pl.DataFrame = None, may_inplace=True):
    ...         if dataframe is None:
    ...             return self._data
    ...         if may_inplace:
    ...             self._data = dataframe
    ...             return self
    ...         else:
    ...             return MyTable(dataframe)

    >>> df = pl.DataFrame({'a': [1, 2, 3], 'b': [10, 20, 30]})
    >>> t = MyTable(df)
    >>> t = t.filter(pl.col("a") > 1).rename({"b": "B"})
    >>> print(t.dataframe())
    shape: (2, 2)
    ┌─────┬─────┐
    │ a   │ B   │
    ├─────┼─────┤
    │ 2   │ 20  │
    │ 3   │ 30  │
    └─────┴─────┘

    Notes
    -----
    - All supported operations delegate to the underlying `polars.DataFrame` and return
      the modified wrapper instance.
    - The actual `dataframe` storage and logic is delegated to subclasses via the abstract
      `dataframe()` getter/setter method.
    - This class is designed for flexible and extensible use in applications such as
      data modeling, pipelines, or typed schema handling.

    Supported Operations
    --------------------
    - Accessors: `columns`, `schema`, `__len__`, `__array__`, `__dataframe__`
    - Indexing: `__getitem__`
    - Structure: `filter`, `drop`, `drop_nulls`, `fill_null`, `fill_nan`, `select`,
                 `with_columns`, `with_row_index`, `rename`, `slice`, `head`, `tail`, `limit`, `sort`
    - Aggregation: `group_by`
    - Partitioning: `partition_by`
    - Joining: `join`
    - Transformation: `pipe`, `clone`, `lazy`

    See Also
    --------
    polars.DataFrame : The underlying DataFrame API
    polars.Expr : Expression system used throughout the API
    """

    @overload
    def dataframe(self) -> pl.DataFrame:
        pass

    @overload
    def dataframe(self, dataframe: pl.DataFrame, may_inplace=True) -> Self:
        pass

    @abc.abstractmethod
    def dataframe(self, dataframe: pl.DataFrame = None, may_inplace=True):
        """
        Getter/setter for the internal Polars DataFrame.

        :param dataframe: Optional new dataframe to set.
        :param may_inplace: If True, update current instance. Otherwise, return new instance.
        :return: The current dataframe or a modified wrapper instance.
        """
        pass

    def __len__(self) -> int:
        """See `polars.DataFrame.__len__`."""
        return len(self.dataframe())

    @property
    def columns(self) -> list[str]:
        """See `polars.DataFrame.columns`."""
        return self.dataframe().columns

    @property
    def schema(self) -> pl.Schema:
        """See `polars.DataFrame.schema`."""
        return self.dataframe().schema

    def __array__(self, *args, **kwargs) -> np.ndarray:
        """See `polars.DataFrame.__array__`."""
        return self.dataframe().__array__(*args, **kwargs)

    def __dataframe__(self, *args, **kwargs):
        """See `polars.DataFrame.__dataframe__`."""
        return self.dataframe().__dataframe__(*args, **kwargs)

    @overload
    def __getitem__(self, key: (
            str
            | tuple[[pty.MultiIndexSelector, pty.SingleColSelector]]
    )) -> pl.Series:
        ...

    @overload
    def __getitem__(self, key: (
            pty.SingleIndexSelector
            | pty.MultiIndexSelector
            | pty.MultiColSelector
            | tuple[pty.SingleIndexSelector, pty.MultiColSelector]
            | tuple[pty.MultiIndexSelector, pty.MultiColSelector]
    )) -> pl.DataFrame:
        ...

    def __getitem__(self, item):
        """See `polars.DataFrame.__getitem__`."""
        return self.dataframe().__getitem__(item)

    def lazy(self) -> LazyDataFrameWrapper[Self]:
        """Wrap dataframe in a lazy wrapper."""
        return LazyDataFrameWrapper(self, self.dataframe().lazy())

    def rename(self, mapping: dict[str, str] | Callable[[str], str]) -> Self:
        """See `polars.DataFrame.rename`."""
        return self.dataframe(self.dataframe().rename(mapping))

    def filter(self, *predicates: pty.IntoExprColumn | Iterable[pty.IntoExprColumn] | bool | list[bool] | np.ndarray,
               **constraints: Any) -> Self:
        """See `polars.DataFrame.filter`."""
        return self.dataframe(self.dataframe().filter(*predicates, **constraints))

    def slice(self, offset: int, length: int | None = None) -> Self:
        """See `polars.DataFrame.slice`."""
        return self.dataframe(self.dataframe().slice(offset, length))

    def head(self, n: int = 5) -> Self:
        """See `polars.DataFrame.head`."""
        return self.dataframe(self.dataframe().head(n))

    def tail(self, n: int = 5) -> Self:
        """See `polars.DataFrame.tail`."""
        return self.dataframe(self.dataframe().tail(n))

    def limit(self, n: int = 5) -> Self:
        """See `polars.DataFrame.limit`."""
        return self.dataframe(self.dataframe().limit(n))

    @overload
    def sort(self,
             by: pty.IntoExpr | Iterable[[pty.IntoExpr]],
             *more_by: pty.IntoExpr,
             descending: bool | Sequence[bool] = False,
             nulls_last: bool | Sequence[bool] = False,
             multithreaded: bool = True,
             maintain_order: bool = False) -> Self:
        ...

    def sort(self, by, *more_by, **kwargs) -> Self:
        """See `polars.DataFrame.sort`."""
        return self.dataframe(self.dataframe().sort(by, *more_by, **kwargs))

    def drop(self, *columns: pty.ColumnNameOrSelector | Iterable[pty.ColumnNameOrSelector],
             strict: bool = True) -> Self:
        """See `polars.DataFrame.drop`."""
        return self.dataframe(self.dataframe().drop(*columns, strict=strict))

    def drop_nulls(self, subset: pty.ColumnNameOrSelector | Collection[pty.ColumnNameOrSelector]) -> Self:
        """See `polars.DataFrame.drop_nulls`."""
        return self.dataframe(self.dataframe().drop_nulls(subset))

    def fill_null(self, value: Any | pl.Expr | None = None,
                  strategy: pty.FillNullStrategy | None = None,
                  limit: int | None = None, **kwargs) -> Self:
        """See `polars.DataFrame.fill_null`."""
        return self.dataframe(self.dataframe().fill_null(value, strategy, limit, **kwargs))

    def fill_nan(self, value: pl.Expr | int | float | None = None) -> Self:
        """See `polars.DataFrame.fill_nan`."""
        return self.dataframe(self.dataframe().fill_nan(value))

    def clear(self, n: int = 5) -> Self:
        """See `polars.DataFrame.clear`."""
        return self.dataframe(self.dataframe().clear(n))

    def clone(self) -> Self:
        """Clone the wrapper."""
        return self.dataframe(self.dataframe(), may_inplace=False)

    @overload
    def partition_by(self, by: pty.ColumnNameOrSelector | Iterable[pty.ColumnNameOrSelector],
                     *more_by: pty.ColumnNameOrSelector,
                     maintain_order: bool = True,
                     include_key: bool = True,
                     as_dict: Literal[False] = ...) -> list[Self]:
        ...

    @overload
    def partition_by(self, by: pty.ColumnNameOrSelector | Iterable[pty.ColumnNameOrSelector],
                     *more_by: pty.ColumnNameOrSelector,
                     maintain_order: bool = ...,
                     include_key: bool = ...,
                     as_dict: Literal[True]) -> dict[tuple[object, ...], Self]:
        ...

    def partition_by(self, by, *more_by, as_dict=False, **kwargs):
        """See `polars.DataFrame.partition_by`."""
        dataframe = self.dataframe().partition_by(by, *more_by, as_dict=as_dict, **kwargs)
        if as_dict:
            return {k: self.dataframe(it, may_inplace=False) for k, it in dataframe.items()}
        else:
            return [self.dataframe(it, may_inplace=False) for it in dataframe]

    def select(self, *exprs: pty.IntoExpr | Iterable[pty.IntoExpr],
               **named_exprs: pty.IntoExpr) -> Self:
        """See `polars.DataFrame.select`."""
        return self.dataframe(self.dataframe().select(*exprs, **named_exprs))

    def with_columns(self, *exprs: pty.IntoExpr | Iterable[pty.IntoExpr],
                     **named_exprs: pty.IntoExpr) -> Self:
        """See `polars.DataFrame.with_columns`."""
        return self.dataframe(self.dataframe().with_columns(*exprs, **named_exprs))

    def with_row_index(self, name: str = "index", offset: int = 0) -> Self:
        """See `polars.DataFrame.with_row_index`."""
        return self.dataframe(self.dataframe().with_row_index(name, offset))

    @overload
    def join(self, other: pl.DataFrame | DataFrameWrapper,
             on: str | pl.Expr | Sequence[str | pl.Expr],
             how: pty.JoinStrategy = "inner", *,
             left_on=None,
             right_on=None,
             suffix: str = "_right",
             validate: pty.JoinValidation = "m:m",
             join_nulls: bool = False,
             coalesce: bool | None = None) -> Self:
        ...

    def join(self, other: pl.DataFrame | DataFrameWrapper, on, *args, **kwargs) -> Self:
        """See `polars.DataFrame.join`."""
        if isinstance(other, DataFrameWrapper):
            other = other.dataframe()
        return self.dataframe(self.dataframe().join(other, on, *args, **kwargs))

    def pipe(self, function: Callable[Concatenate[pl.DataFrame, P], pl.DataFrame],
             *args: P.args,
             **kwargs: P.kwargs) -> Self:
        """See `polars.DataFrame.pipe`."""
        return self.dataframe(self.dataframe().pipe(function, *args, **kwargs))

    def group_by(self, *by: pty.IntoExpr | Iterable[pty.IntoExpr],
                 maintain_order: bool = False,
                 **named_by: pty.IntoExpr) -> GroupBy:
        """See `polars.DataFrame.group_by`."""
        return self.dataframe().group_by(*by, maintain_order=maintain_order, **named_by)


T = TypeVar('T', bound=DataFrameWrapper)


class LazyDataFrameWrapper(Generic[T]):
    __slots__ = '__wrapper', '__lazy'

    def __init__(self, wrapper: T, lazy: pl.LazyFrame):
        self.__wrapper = wrapper
        self.__lazy = lazy

    @property
    def columns(self) -> list[str]:
        return self.__lazy.columns

    @property
    def schema(self) -> pl.Schema:
        return self.__lazy.schema

    def lazy(self) -> LazyDataFrameWrapper[T]:
        return self

    def collect(self, **kwargs) -> T:
        return self.__wrapper.dataframe(self.__lazy.collect(**kwargs))

    def rename(self, mapping: dict[str, str] | Callable[[str], str]) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.rename(mapping))

    def slice(self, offset: int, length: int | None = None) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.slice(offset, length))

    def head(self, n: int = 5) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.head(n))

    def tail(self, n: int = 5) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.tail(n))

    def limit(self, n: int = 5) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.limit(n))

    def clear(self, n: int = 0) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.clear(n))

    def filter(self, *predicates: pty.IntoExprColumn | Iterable[pty.IntoExprColumn] | bool | list[bool] | np.ndarray,
               **constraints: Any) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.filter(*predicates, **constraints))

    @overload
    def sort(self,
             by: pty.IntoExpr | Iterable[[pty.IntoExpr]],
             *more_by: pty.IntoExpr,
             descending: bool | Sequence[bool] = False,
             nulls_last: bool | Sequence[bool] = False,
             multithreaded: bool = True,
             maintain_order: bool = False) -> Self:
        pass

    def sort(self, by, *more_by, **kwargs) -> Self:
        df = self.__lazy.sort(by, *more_by, **kwargs)
        return LazyDataFrameWrapper(self.__wrapper, df)

    def drop(self, *columns: pty.ColumnNameOrSelector | Iterable[pty.ColumnNameOrSelector],
             strict: bool = True) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.drop(*columns, strict=strict))

    def drop_nulls(self, subset: pty.ColumnNameOrSelector | Collection[pty.ColumnNameOrSelector]) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.drop_nulls(subset))

    def fill_null(self, value: Any | pl.Expr | None = None,
                  strategy: pty.FillNullStrategy | None = None,
                  limit: int | None = None, **kwargs) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.fill_null(value, strategy, limit, **kwargs))

    def fill_nan(self, value: pl.Expr | int | float | None = None) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.fill_nan(value))

    def select(self, *exprs: pty.IntoExpr | Iterable[pty.IntoExpr],
               **named_exprs: pty.IntoExpr) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.select(*exprs, **named_exprs))

    def with_columns(self, *exprs: pty.IntoExpr | Iterable[pty.IntoExpr],
                     **named_exprs: pty.IntoExpr) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.with_columns(*exprs, **named_exprs))

    def with_row_index(self, name: str = "index", offset: int = 0) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.with_row_index(name, offset))

    @overload
    def join(self, other: pl.DataFrame | DataFrameWrapper,
             on: str | pl.Expr | Sequence[str | pl.Expr],
             how: pty.JoinStrategy = "inner", *,
             left_on=None,
             right_on=None,
             suffix: str = "_right",
             validate: pty.JoinValidation = "m:m",
             join_nulls: bool = False,
             coalesce: bool | None = None) -> Self:
        pass

    def join(self, other: pl.DataFrame | DataFrameWrapper, on, *args, **kwargs) -> Self:
        if isinstance(other, DataFrameWrapper):
            other = other.dataframe()
        if not isinstance(other, pl.LazyFrame):
            other = other.lazy()

        df = self.__lazy.join(other, on, *args, **kwargs)
        return LazyDataFrameWrapper(self.__wrapper, df)

    def pipe(self, function: Callable[Concatenate[pl.LazyFrame, P], pl.LazyFrame],
             *args: P.args,
             **kwargs: P.kwargs) -> Self:
        return LazyDataFrameWrapper(self.__wrapper, self.__lazy.pipe(function, *args, **kwargs))


def helper_with_index_column(df: T,
                             column: str,
                             index: int | list[int] | np.ndarray | T,
                             maintain_order: bool = False,
                             strict: bool = False) -> T:
    """
    A help function to do the filter on an index column.

    :param df:
    :param column: index column
    :param index: index array
    :param maintain_order: keep the ordering of *index* in the returned dataframe.
    :param strict: all index in *index* should present in the returned dataframe. Otherwise, an error will be raised.
    :return:
    :raise RuntimeError: strict mode fail.
    """
    if isinstance(index, (int, np.integer)):
        ret = df.filter(pl.col(column) == index)
        if len(ret) == 0 and strict:
            raise RuntimeError(f'missing {column}: [{index}]')
    elif isinstance(index, type(df)):
        index = index[column].to_numpy()
    else:
        index = np.asarray(index)

    if strict:
        if len(miss := np.setdiff1d(index, df.dataframe()[column].unique().to_numpy())) > 0:
            raise RuntimeError(f'missing {column}: {list(miss)}')

    if maintain_order:
        _column = '_' + column
        index = pl.DataFrame(
            {column: index},
            schema_overrides={column: df.schema[column]}
        ).with_row_index(_column)
        ret = df.lazy().join(index, on=column, how='left')
        ret = ret.filter(pl.col(_column).is_not_null())
        return ret.sort(_column).drop(_column).collect()
    else:
        return df.filter(pl.col(column).is_in(index))


def assert_polars_equal_verbose(df1: pl.DataFrame, df2: pl.DataFrame, **kwargs):
    """
    Assert that two Polars DataFrames are equal and provide detailed diagnostics if they differ

    :param df1: The first Polars DataFrame to compare
    :param df2: The second Polars DataFrame to compare
    :param kwargs: keyword arguments passed to :func:`~neuralib.util.verbose.printdf()`
    :return:
    """
    try:
        assert_frame_equal(df1, df2)
        print('DataFrames are equal.')
    except AssertionError as e:
        print('DataFrames are NOT equal.')

        # shape
        print('\nShape mismatch:')
        print(f'df1: {df1.shape}')
        print(f'df2: {df2.shape}')

        # column
        if df1.columns != df2.columns:
            print('\nColumn mismatch:')
            print(f'df1 columns: {df1.columns}')
            print(f'df2 columns: {df2.columns}')
            raise e

        df1_extra = df1.join(df2, on=df1.columns, how='anti')
        df2_extra = df2.join(df1, on=df1.columns, how='anti')

        if df1_extra.height > 0:
            print('\nRows in df1 not in df2:')
            printdf(df1_extra, **kwargs)

        if df2_extra.height > 0:
            print('\nRows in df2 not in df1:')
            printdf(df2_extra, **kwargs)

        # If shapes match, show cell-wise diff
        if df1.shape == df2.shape:
            print('\nCell-wise differences (non-equal values):')
            diffs = _highlight_cell_differences(df1, df2)
            printdf(diffs, **kwargs)

        raise e


def _highlight_cell_differences(df1: pl.DataFrame, df2: pl.DataFrame) -> pl.DataFrame:
    return pl.DataFrame({
        col: df1[col].cast(str).zip_with(
            df1[col] != df2[col],
            pl.lit('df1=') + df1[col].cast(str) + ', df2=' + df2[col].cast(str)
        ).fill_null('')  # Handle NaNs
        for col in df1.columns
    })



from pathlib import Path
from typing import Literal, NamedTuple, Iterable

import numpy as np
import pandas as pd
import polars as pl

from neuralib.atlas.map import NUM_MERGE_LAYER
from neuralib.atlas.typing import Source, HEMISPHERE_TYPE
from neuralib.typing import DataFrame

__all__ = [
    'ALLEN_CCF_10um_BREGMA',
    #
    'SourceCoordinates',
    'iter_source_coordinates',
    'get_margin_merge_level',
    'allen_to_brainrender_coord',
    'as_coords_array',
]

ALLEN_CCF_10um_BREGMA = np.array([540, 0, 570])  # AP, DV, LR
"""allen CCF 10um volume coordinates, refer to allenCCF/Browsing Functions/allenCCFbregma.m"""


class SourceCoordinates(NamedTuple):
    source: Source
    coordinates: np.ndarray
    """AP, DV, ML coordinates. `Array[float, [N, 3]]`"""

    @property
    def ap(self) -> np.ndarray:
        return self.coordinates[:, 0]

    @property
    def dv(self) -> np.ndarray:
        return self.coordinates[:, 1]

    @property
    def ml(self) -> np.ndarray:
        return self.coordinates[:, 2]


def iter_source_coordinates(
        file: Path,
        *,
        area: list[str] | str | None = None,
        source: list[Source] | Source | None = None,
        region_col: str | None = None,
        hemisphere: HEMISPHERE_TYPE = 'both',
        to_brainrender: bool = True,
        source_order: tuple[Source, ...] | None = None,
        inverse_hemisphere: bool = False,
) -> Iterable[SourceCoordinates]:
    """Load allen ccf roi output (merged different color channels).

    :param file: parsed csv file after
    :param area: only show rois in region(s)
    :param source: only show rois from source(s)
    :param region_col: if None, auto infer, and check the lowest merge level contain all the regions specified
    :param hemisphere: which brain hemisphere
    :param to_brainrender: convert the coordinates to brain render
    :param source_order: whether specify the source generator order
    :param inverse_hemisphere: if True, then inverse ML coordinates. default is False.
    :return: Iterable of :class:`SourceCoordinates`
    """
    df = pl.read_csv(file)
    #
    if area is not None and len(area) != 0:
        if isinstance(area, str):
            area = [area]

        if region_col is None:
            region_col = get_margin_merge_level(df, area, 'lowest')

        df = df.filter(pl.col(region_col).is_in(area))
        if df.is_empty():
            raise RuntimeError('check lowest merge level')

    #
    if source is not None and len(source) != 0:
        if isinstance(source, str):
            source = [source]

        df = df.filter(pl.col('source').is_in(source))
        if df.is_empty():
            raise RuntimeError(f'empty df: likely incorrect source selected {source}')

    #
    if hemisphere != 'both':
        df = df.filter(pl.col('hemisphere') == hemisphere)

    #
    coords = as_coords_array(df)

    if inverse_hemisphere:
        coords[:, 2] *= -1

    if to_brainrender:
        coords = allen_to_brainrender_coord(coords)

    if source_order is None:
        source_order = df['source'].unique()

    for src in source_order:
        mx = df.select(pl.col('source') == src).to_numpy()[:, 0]
        yield SourceCoordinates(src, coords[mx])


def get_margin_merge_level(df: pl.DataFrame,
                           areas: list[str] | str,
                           margin: Literal['lowest', 'highest']) -> str:
    """Get the lowest or highest merge level (i.e., parsed_csv) containing all the regions

    :param df: parsed csv
    :param areas: an area or a list of areas
    :param margin: get the either lowest of highest merge level for a given area
    :return: col name if parsed csv
    """
    if not isinstance(areas, (tuple, list)):
        areas = [areas]

    eval_merge = [f'tree_{i}' for i in range(NUM_MERGE_LAYER)]

    if margin == 'lowest':
        level = eval_merge
    elif margin == 'highest':
        level = reversed(eval_merge)
    else:
        raise ValueError('')

    for lv in level:
        ls = df[lv]
        if np.all([a in ls for a in areas]):
            return lv

    raise ValueError(f'{areas} not found')


def allen_to_brainrender_coord(data: DataFrame | np.ndarray) -> np.ndarray:
    """Convert coordinates space of ``AllenCCF`` to ``Brainrender`` coordinates

    :param data: Dataframe with 'AP_location', 'DV_location', 'ML_location' headers.
            Or numpy array with `Array[float, [N, 3]]` or `Array[float, 3]`
    :return: brainrender coordinates. `Array[float, [N, 3]]` with AP, DV, ML coordinates
    """
    coords = as_coords_array(data)

    coords *= 1000
    coords[:, 0] /= -1  # increment toward posterior
    coords[:, 2] /= -1  # increment toward left hemisphere
    bregma = ALLEN_CCF_10um_BREGMA * 10  # pixel to um
    coords += bregma  # roi relative to bregma

    return coords


def as_coords_array(data: DataFrame | np.ndarray) -> np.ndarray:
    """
    Convert dataframe/1D numpy array to coordinates numpy array

    :param data: Dataframe with 'AP_location', 'DV_location', 'ML_location' headers.
        Or numpy array with `Array[float, [N, 3]]` or `Array[float, 3]`
    :return: `Array[float, [N, 3]]` with AP, DV, ML coordinates
    """
    match data:
        case pd.DataFrame():
            coords = pl.from_pandas(data).select('AP_location', 'DV_location', 'ML_location').to_numpy()
        case pl.DataFrame():
            coords = data.select('AP_location', 'DV_location', 'ML_location').to_numpy()
        case np.ndarray():
            if data.ndim == 1:
                data = np.expand_dims(data, axis=0)
            if data.ndim != 2 or data.shape[1] != 3:
                raise ValueError(f'{data.ndim=}, {data.shape=}')
            coords = data
        case _:
            raise TypeError('')

    return coords



from typing import NamedTuple

import numpy as np
from scipy import stats
from scipy.interpolate import interp1d
from typing_extensions import Self

from neuralib.util.segments import segment_epochs, segment_duration, segment_contains
from neuralib.util.unstable import unstable

__all__ = [
    'CircularPosition',
    'interp_pos1d',
    #
    'speed_2d',
    'direction_2d',
    'interp_gap2d'
]


class CircularPosition(NamedTuple):
    """
    Position information in circular environment.

    `Dimension parameters`:

        P = Number of position points (after interpolation)

        T = Number of trial(lap)

    """

    t: np.ndarray
    """1D time array in s. `Array[float, P]`"""

    p: np.ndarray
    """1D position array in cm. `Array[float, P]`"""

    d: np.ndarray
    """1D displacement array in cm. `Array[float, P]`"""

    v: np.ndarray
    """1D velocity array in cm/s. `Array[float, P]`"""

    trial_time_index: np.ndarray
    """1D time index for every trial. `Array[int, T]`"""

    def with_time_range(self, t0: float, t1: float) -> Self:
        """With specific time range

        :param t0: time start
        :param t1: time end
        :return: A new instance of the class with updated subset of attributes (t, p, d, v) based on the time range.
        """
        tx0 = self.t >= t0
        tx1 = self.t <= t1
        tx = np.logical_and(tx0, tx1)

        ix0 = np.nonzero(tx0)[0][0]  # start index for time
        ix1 = np.nonzero(~tx1)[0][0]  # end index for time
        ix = np.logical_and(self.trial_time_index >= ix0, self.trial_time_index <= ix1)

        return self._replace(
            t=self.t[tx],
            p=self.p[tx],
            d=self.d[tx],
            v=self.v[tx],
            trial_time_index=self.trial_time_index[ix]
        )

    def with_run_mask1d(self, **kwargs) -> Self:
        """With only the running epoch. **Note that shape will be changed (stationary epoch excluded)**

        :param kwargs: Additional keyword arguments to pass to running_mask1d function.
        :return: A new instance of the class with updated subset of attributes (t, p, d, v) based on the mask generated.
        """
        from .epoch import running_mask1d
        x = running_mask1d(self.t, self.v, **kwargs)
        return self._replace(
            t=self.t[x],
            p=self.p[x],
            d=self.d[x],
            v=self.v[x],
        )

    def interp_time(self, time: np.ndarray, fill_value: float = 0.0) -> Self:
        """Interpolate position data to new time points.
        
        This method interpolates the position (p), displacement (d), and velocity (v)
        data to match the provided time array. Uses linear interpolation and fills
        out-of-bounds values with the specified fill_value.
        
        :param time: Target time array to interpolate to. `Array[float, N]`
        :param fill_value: Value to use for out-of-bounds interpolation, defaults to 0.0.
        :return: A new CircularPosition instance with interpolated data aligned to the new time array.
        """
        p = interp1d(self.t, self.p, bounds_error=False, fill_value=fill_value)(time)
        d = interp1d(self.t, self.d, bounds_error=False, fill_value=fill_value)(time)
        v = interp1d(self.t, self.v, bounds_error=False, fill_value=fill_value)(time)

        return self._replace(t=time, p=p, d=d, v=v)

    @property
    def trial_array(self) -> np.ndarray:
        """Trial number array as same shape as *P*

        :return: `Array[int, P]`
        """
        ret = np.zeros_like(self.t, int)
        ret[self.trial_time_index] = 1
        return np.cumsum(ret)


def interp_pos1d(time: np.ndarray,
                 pos: np.ndarray,
                 *,
                 sampling_rate: float = 1000,
                 norm_max_value: float = 150,
                 remove_nan: bool = True,
                 renew_trial_value: float = -100) -> CircularPosition:
    """
    Interpolate the raw position data

    `Dimension parameters`:

        P = Number of position points (Raw)

    :param time: An array of time stamps corresponding to positional data. `Array[float, P]`.
    :param pos: An array of positional data (a.u, encoder or other hardware readout). `Array[float, P]`.
    :param sampling_rate: The rate at which data points should be sampled, defaults to 1000.
    :param norm_max_value: The value by which to normalize the positional data, defaults to 150 (cm).
    :param remove_nan: Flag indicating whether to remove NaN values from interpolated data, defaults to True.
    :param renew_trial_value: The value used to detect the overflow index which signifies trial renewals, defaults to -100.
    :return: ``CircularPosition``
    """

    overflow_idx = np.nonzero(np.diff(pos) < renew_trial_value)[0]

    # Get the position(p), displacement(d)
    if len(overflow_idx) == 0:  # not more than a trial
        p = pos / np.max(pos)
        d = pos / np.max(pos)
    else:
        p = np.zeros_like(pos, float)
        d = np.zeros_like(pos, float)
        mode = stats.mode(pos[overflow_idx], keepdims=True).mode.astype(int)

        # fix the first trial
        p[:overflow_idx[0] + 1] = pos[:overflow_idx[0] + 1] / pos[overflow_idx[0]]
        d[:overflow_idx[0] + 1] = pos[:overflow_idx[0] + 1] / pos[overflow_idx[0]]
        lap_counter = 1

        # loop through rest of trials
        for lap_start, lap_stop in zip(overflow_idx[:-1], overflow_idx[1:]):
            p[lap_start + 1:lap_stop + 1] = pos[lap_start + 1:lap_stop + 1] / pos[lap_stop]
            d[lap_start + 1:lap_stop + 1] = pos[lap_start + 1:lap_stop + 1] / pos[lap_stop] + lap_counter
            lap_counter += 1

        # fix the last lap
        p[overflow_idx[-1] + 1:] = p[overflow_idx[-1] + 1:] / mode
        d[overflow_idx[-1] + 1:] = p[overflow_idx[-1] + 1:] / mode + lap_counter

    # actual length
    p *= norm_max_value
    d *= norm_max_value

    # interpolate
    t0 = np.min(time)
    t1 = np.max(time)
    tt = np.linspace(t0, t1, int((t1 - t0) * sampling_rate))
    pp = interp1d(time, p, kind='nearest', copy=False, bounds_error=False, fill_value=np.nan)(tt)
    dd = interp1d(time, d, kind='linear', copy=False, bounds_error=False, fill_value='extrapolate')(tt)
    vv = np.diff(dd, prepend=dd[0]) * sampling_rate

    # remove nan from position_value (from interpolation boundary error)
    if remove_nan:
        x = ~np.isnan(pp)
        tt = tt[x]
        pp = pp[x]
        dd = dd[x]
        vv = vv[x]

    # time index foreach trial
    if len(overflow_idx) == 0:
        trial_time_index = np.array([0])
    else:
        trial_time_index = np.zeros_like(overflow_idx, int)
        for i, t in enumerate(time[overflow_idx]):
            # shape of t changed after interpolation(tt), find the minimal tt right after the t
            trial_time_index[i] = np.nonzero(tt >= t)[0][0]

    return CircularPosition(tt, pp, dd, vv, trial_time_index)


# =========== #
# 2D Movement #
# =========== #

def _complex_2d(xy: np.ndarray, dt: float = 1.0) -> np.ndarray:
    return np.gradient(xy[:, 0], dt) + np.gradient(xy[:, 1], dt) * 1.0j


def speed_2d(xy: np.ndarray, dt: float = 1.0) -> np.ndarray:
    """
    Compute speed 2D movement

    :param xy: 2D array coordinates in xy. `Array[float, [N, 2]]`
    :param dt: Time period between samples (for smoothing)
    :return: Speed. `Array[float, N]`
    """
    return np.abs(_complex_2d(xy, dt))


def direction_2d(xy: np.ndarray, dt: float = 1.0) -> np.ndarray:
    """
    Compute direction 2D movement

    :param xy: 2D array coordinates in xy. `Array[float, [N, 2]]`
    :param dt: Time period between samples (for smoothing)
    :return: Movement direction. `Array[float, N]`
    """
    return np.angle(_complex_2d(xy, dt), deg=True)


@unstable()
def interp_gap2d(t: np.ndarray,
                 xy: np.ndarray,
                 duration: float) -> np.ndarray:
    """
    Interpolate over gaps in position record.

    :param t: Time Vector. `Array[float, N]`
    :param xy: 2D array coordinates in xy. `Array[float, [N, 2]]`
    :param duration: Maximum duration of a gap that should be interpolated over. same unit as ``t``
    :return: Corrected x,y coordinates. `Array[float, [N, 2]]`
    """
    invalid = np.isnan(xy[:, 0])
    valid = ~invalid

    # select small gaps
    invalid_seg = segment_epochs(invalid, t)
    invalid_seg = invalid_seg[segment_duration(invalid_seg) < duration]

    if invalid_seg.shape[0] > 0:
        print(f'{len(invalid_seg)} segments invalid')
        invalid_indices = np.nonzero(segment_contains(invalid_seg, t))[0]

        f = interp1d(
            t[valid],
            xy[valid, :],
            kind="linear",
            bounds_error=False,
            axis=0,
            assume_sorted=True,
        )
        xy[invalid_indices, :] = f(t[invalid_indices])

    return xy



import functools
import warnings
from typing import ParamSpec
from typing import TypeVar, Callable, Type

__all__ = ['unstable']

P = ParamSpec("P")
R = TypeVar("R")
F = TypeVar('F', bound=Callable[P, R])
T = TypeVar('T', Type[R], F)


def unstable(doc: bool = True,
             runtime: bool = True,
             mark_all: bool = False,
             method: str = '__init__') -> Callable[[T], T]:
    """
    A decorator that mark class/function unstable.

    :param doc: Add unstable message in function/class document.
    :param runtime: Add runtime warning when invoking function/initialization.
    :param mark_all: Mark all public methods when decorate on a class, If False, then only mark ``method_name``
    :param method: Name of method in class to be marked. defaults to '__init__'
    """

    def _decorator(obj: T) -> T:
        if getattr(obj, '__unstable_marker', False):
            return obj

        try:
            obj.__unstable_marker = True
        except AttributeError:
            # AttributeError: 'wrapper_descriptor' object has no attribute
            pass

        if doc:
            new_doc = 'UNSTABLE.'
            if obj.__doc__ is not None:
                new_doc += f'\n{obj.__doc__}'

            try:
                obj.__doc__ = new_doc
            except AttributeError:
                # AttributeError: 'wrapper_descriptor' object has no attribute
                pass

        if isinstance(obj, type):  # class
            if mark_all:
                for meth in dir(obj):
                    if callable(f := getattr(obj, meth, None)) and (not meth.startswith('_') or meth in ('__init__',)):
                        setattr(obj, meth, unstable(doc=doc, runtime=runtime, mark_all=mark_all)(f))
            else:
                f = obj.__dict__.get(method)
                if f is not None:
                    setattr(obj, method, unstable(doc=doc, runtime=runtime, mark_all=mark_all)(f))
                else:
                    raise RuntimeError(f'Method {method} in {obj.__qualname__} not found')

            return obj

        else:  # func/meth
            if runtime:
                @functools.wraps(obj)
                def _unstable_meth(*args: P.args, **kwargs: P.kwargs) -> R:
                    warnings.warn(f"{obj.__qualname__} is unstable and under development.", stacklevel=2)
                    return obj(*args, **kwargs)

                return _unstable_meth
            else:
                return obj

    return _decorator



from __future__ import annotations

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.patches import Patch
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from scipy.interpolate import interp1d
from typing import NamedTuple, Literal
from typing_extensions import Self

from neuralib.model.rastermap import RasterMapResult
from neuralib.plot import plot_figure
from neuralib.typing import PathLike
from neuralib.util.deprecation import deprecated_class

__all__ = [
    'plot_rastermap',
    'plot_cellular_spatial',
    'plot_wfield_spatial',
    'RasterMapPlot',
    'Covariant'
]


def plot_rastermap(result: RasterMapResult,
                   act_time: np.ndarray, *,
                   time_range: tuple[float, float] | None = None,
                   covars: list[Covariant] | None = None,
                   output: PathLike | None = None):
    """
    plot the rastermap result with behavioral measurements

    :param result: :class:`~.core.RasterMapResult`
    :param act_time: neural activity time array. should be the same T as neural_activity when run the rastermap
    :param time_range: time range for plotting (START,END)
    :param covars: list of :class:`~Covariant`
    :param output: output path for figure save. If None then show
    """
    plotter = RasterMapPlot(result, act_time, time_range, covars, output)
    plotter.plot_rastermap()


def plot_cellular_spatial(result: RasterMapResult,
                          xpos: np.ndarray,
                          ypos: np.ndarray,
                          ax: Axes | None = None,
                          output: PathLike | None = None,
                          **kwargs):
    """
    Plot spatial location of each cell cluster by rastermap

    :param result: :class:`~.core.RasterMapResult`
    :param xpos: soma central X position.`Array[float, N]`
    :param ypos: soma central Y position.`Array[float, N]`
    :param ax: ``Axes``
    :param output: output path for figure save. If None then show
    :param kwargs: additional arguments pass to ``ax.set()``
    :return:
    """
    if ax is None:
        _, ax = plt.subplots()

    ax.scatter(xpos, ypos, s=8, c=result.embedding, cmap="gist_ncar", alpha=0.25)
    ax.invert_yaxis()
    ax.set(**kwargs)
    ax.set_aspect('equal')

    if output is not None:
        plt.savefig(output)
    else:
        plt.show()


def plot_wfield_spatial(result: RasterMapResult,
                        width: int,
                        height: int,
                        ax: Axes | None = None,
                        output: PathLike | None = None,
                        **kwargs):
    """
    Plot spatial location of each pixel cluster by rastermap

    :param result: :class:`~.core.RasterMapResult`
    :param width: sequence image width
    :param height: sequence image height
    :param ax: ``Axes``
    :param output: output path for figure save. If None then show
    :param kwargs: additional arguments pass to ``ax.set()``
    """

    if ax is None:
        _, ax = plt.subplots()

    x = np.arange(width)
    y = np.arange(height)
    xpos, ypos = np.meshgrid(x, y)  # Array[float, [W, H]]

    ax.scatter(xpos, ypos, s=1, c=result.embedding, cmap="gist_ncar", alpha=0.25)
    ax.invert_yaxis()
    ax.set(**kwargs)
    ax.set_aspect('equal')

    if output is not None:
        plt.savefig(output)
    else:
        plt.show()


class RasterMapPlot:
    """Plot the rastermap result with behavioral measurements"""

    DEFAULT_TRIAL_COLORS = ('royalblue', 'darkorange', 'seagreen', 'crimson')

    def __init__(self, result: RasterMapResult,
                 act_time: np.ndarray,
                 time_range: tuple[float, float] | None = None,
                 covars: list[Covariant] | None = None,
                 event_colors: list[str] | None = None,
                 output: PathLike | None = None):
        """
        :param result: class:`~.core.RasterMapResult`
        :param act_time: neural activity time array. should be the same T as neural_activity when run the rastermap
        :param time_range: time range for plotting (START,END)
        :param covars: list of :class:`~Covariant`
        """
        self.raster = result
        self.covars = covars

        self.time_range = time_range or (act_time[0], act_time[-1])

        if time_range is not None:
            self.act_mask = np.logical_and(time_range[0] < act_time, act_time < time_range[1])
        else:
            self.act_mask = np.ones_like(act_time, dtype=np.bool_)

        self.act_time = act_time[self.act_mask]

        self.event_colors = event_colors or self.DEFAULT_TRIAL_COLORS
        self.output = output

    @property
    def super_neurons(self) -> np.ndarray:
        """rastermap sorted 2D array. `Array[float, [N, T]]`"""
        return self.raster.super_neurons[:, self.act_mask]

    def process_continuous(self) -> list[Covariant]:
        """process behavioral measurements, select time range and do the interpolation same shape as neural activity"""
        return [
            cov.masking_time(self.time_range).interp_activity(self.act_time)
            for cov in self.covars
            if cov.dtype == 'continuous'
        ]

    def plot_rastermap(self):
        if self.covars is not None:
            covars = self.process_continuous()
            n_covars = len(covars)
        else:
            covars = None
            n_covars = 1

        height_ratios = [1] * n_covars + [7]
        with plot_figure(self.output, n_covars + 1, 1, gridspec_kw={'height_ratios': height_ratios},
                         tight_layout=False, sharex=True) as _ax:

            # continuous dtype
            if self.covars is not None:
                for i, cov in enumerate(covars):
                    if cov.dtype == 'continuous':
                        ax = _ax[i]
                        ax.plot(cov.time, cov.value, color='k')
                        ax.set_xlim(self.time_range)
                        ax.axis('off')
                        ax.set_title(cov.name)
            else:
                _ax[0].axis('off')

            # rastermap
            ax = _ax[n_covars]
            ax.imshow(
                self.super_neurons,
                cmap='gray_r',
                vmin=0,
                vmax=0.8,
                aspect='auto',
                interpolation='none',
                extent=(self.time_range[0], self.time_range[1], self.raster.n_clusters, 0),
            )
            ax.set(xlabel='time', ylabel='rastermap clusters')

            # event segments
            if self.covars is not None:
                self.plot_segments(ax=ax)

            # colormap
            n_clusters = self.raster.n_clusters
            cluster_colors = plt.get_cmap("gist_ncar", n_clusters)
            cb_ax = inset_axes(ax, width="2%", height="100%", loc="right",
                               bbox_to_anchor=(0.05, 0., 1, 1), bbox_transform=ax.transAxes, borderpad=0)

            cb_ax.imshow(
                np.arange(n_clusters)[:, np.newaxis],
                cmap=cluster_colors,
                aspect="auto"
            )
            cb_ax.axis("off")

    def plot_segments(self, ax: Axes):
        """
        Plot event segments as vertical spans on the axis

        :param ax: matplotlib Axes object
        """
        legend_patches = []

        for i, cov in enumerate(self.covars):
            if cov.dtype == 'event':
                for start, end in cov.time:
                    if self.time_range is not None:
                        t0 = self.time_range[0]
                        t1 = self.time_range[1]
                        if end < t0 or start > t1:
                            continue  # completely outside view
                        start = max(start, t0)
                        end = min(end, t1)

                    ax.axvspan(start, end, color=self.event_colors[i], alpha=0.4)

                legend_patches.append(Patch(facecolor=self.event_colors[i], alpha=0.4, label=cov.name))

        if legend_patches:
            ax.legend(handles=legend_patches, loc='upper right', fontsize=10, frameon=False)


class Covariant(NamedTuple):
    """
    Covariant variable that can be plotted alongside rastermap results.

    Supports two types:

    - ``'continuous'``: time-series data (e.g., velocity, position)
    - ``'event'``: discrete time segments (e.g., trial periods, behavioral events)

    :param name: name of the covariant variable
    :param dtype: type of the covariant variable (``'event'`` or ``'continuous'``)
    :param time: time array. ``Array[float, T]`` for continuous dtype or ``Array[float, [E, 2]]`` for on/off event dtype
    :param value: value array (only for continuous dtype). ``Array[float, T]``
    """
    name: str
    dtype: Literal['event', 'continuous']
    time: np.ndarray
    value: np.ndarray | None = None

    def masking_time(self, t: tuple[float, float]) -> Self:
        """
        Mask data to a specific time range (continuous dtype only).

        :param t: (START,END) time range
        :return: new Covariant with data filtered to the time range
        :raises ValueError: if called on event dtype
        """
        if self.dtype == 'event':
            raise ValueError('method only available for continuous dtype')

        mx = np.logical_and(t[0] < self.time, self.time < t[1])
        return self._replace(time=self.time[mx], value=self.value[mx])

    def interp_activity(self, act_time: np.ndarray) -> Self:
        """
        Interpolate data to match another activity time array (continuous dtype only).

        :param act_time: activity time array to interpolate to. `Array[float, T']`
        :return: new Covariant with interpolated values matching act_time
        :raises ValueError: if called on event dtype
        """
        if self.dtype == 'event':
            raise ValueError('method only available for continuous dtype')

        v = interp1d(self.time, self.value, bounds_error=False, fill_value=0)(act_time)
        return self._replace(time=act_time, value=v)


@deprecated_class(new='neuralib.model.rastermap.plot.Covariant', removal_version='0.6.0')
class BehavioralVT(NamedTuple):
    name: str
    """name of the behavioral variable"""
    time: np.ndarray
    """time array. `Array[float, T]`"""
    value: np.ndarray
    """value array. `Array[float, T]`"""

    def masking_time(self, t: tuple[float, float]) -> Self:
        """
        mask given time range
        :param t: (START,END) time range
        :return:
        """
        mx = np.logical_and(t[0] < self.time, self.time < t[1])
        return self._replace(time=self.time[mx], value=self.value[mx])

    def interp_activity(self, act_time: np.ndarray) -> Self:
        """
        interpolation to another activity array. i.e., neural activity
        :param act_time: activity array. `Array[float, T']`
        :return:
        """
        v = interp1d(self.time, self.value, bounds_error=False, fill_value=0)(act_time)
        return self._replace(time=act_time, value=v)



"""
SQL window functions.

https://www.sqlite.org/windowfunctions.html
https://www.sqlite.org/windowfunctions.html#built_in_window_functions
"""

from . import expr
from .expr import SqlWindowFunc
from .func_dec import as_func_expr

__all__ = [
    'window_def', 'row_number', 'rank', 'dense_rank', 'percent_rank', 'cume_dist', 'ntile', 'lag', 'lead',
    'first_value', 'last_value', 'nth_value'
]


def window_def(name: str = None, *, order_by: list = None, partition_by: list = None) -> expr.SqlWindowDef:
    return expr.SqlWindowDef(name).over(order_by=order_by, partition_by=partition_by)


@as_func_expr(func=SqlWindowFunc)
def row_number() -> SqlWindowFunc:
    raise NotImplementedError()


@as_func_expr(func=SqlWindowFunc)
def rank() -> SqlWindowFunc:
    raise NotImplementedError()


@as_func_expr(func=SqlWindowFunc)
def dense_rank() -> SqlWindowFunc:
    raise NotImplementedError()


@as_func_expr(func=SqlWindowFunc)
def percent_rank() -> SqlWindowFunc:
    raise NotImplementedError()


@as_func_expr(func=SqlWindowFunc)
def cume_dist() -> SqlWindowFunc:
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr(func=SqlWindowFunc)
def ntile(n: int) -> SqlWindowFunc:
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr(func=SqlWindowFunc)
def lag(expr, offset: int = None, default=None) -> SqlWindowFunc:
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr(func=SqlWindowFunc)
def lead(expr, offset: int = None, default=None) -> SqlWindowFunc:
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr(func=SqlWindowFunc)
def first_value(expr) -> SqlWindowFunc:
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr(func=SqlWindowFunc)
def last_value(expr) -> SqlWindowFunc:
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr(func=SqlWindowFunc)
def nth_value(expr, n: int) -> SqlWindowFunc:
    raise NotImplementedError()



from __future__ import annotations

from typing import Literal, Sequence

import numpy as np
import polars as pl
from polars.exceptions import ColumnNotFoundError
from typing_extensions import Self

from neuralib.atlas.cellatlas import load_cellatlas
from neuralib.atlas.data import load_bg_volumes
from neuralib.atlas.map import merge_until_level, NUM_MERGE_LAYER, DEFAULT_FAMILY_DICT
from neuralib.atlas.typing import Channel, HEMISPHERE_TYPE
from neuralib.atlas.util import get_margin_merge_level
from neuralib.typing import PathLike
from neuralib.util.dataframe import DataFrameWrapper
from neuralib.util.utils import ensure_dir
from neuralib.util.verbose import print_save, print_load

__all__ = [
    'ROIS_NORM_TYPE',
    'RoiClassifierDataFrame',
    'RoiNormalizedDataFrame',
    'RoiSubregionDataFrame',
]

ROIS_NORM_TYPE = Literal['channel', 'volume', 'cell', 'none']
"""Roi normalized type"""


class RoiClassifierDataFrame(DataFrameWrapper):
    """
    RoiClassifierDataFrame with each roi (rows)

    - Required fields:
        - ``acronym`` - area acronym
        - ``AP_location``: anterior-posterior coordinates (mm)
        - ``DV_location``: dorsal-ventral coordinates (mm)
        - ``ML_location``: medial-lateral coordinates (mm)
        - ``channel``: fluorescence channel (i.e., gfp, rfp, mcherry, ...)
        - ``source``: source name (i.e., if circuit tracing, you can give source tracing area...)

    """
    _required_fields = ('acronym', 'AP_location', 'DV_location', 'ML_location', 'channel', 'source')
    _valid_classified_fields = ('acronym', 'tree_0', 'tree_1', 'tree_2', 'tree_3', 'tree_4', 'family')

    def __init__(self, df: pl.DataFrame, *,
                 cached_dir: PathLike | None = None,
                 invalid_post_processing_cache: bool = False):
        """

        :param df: DataFrame with required fields
        :param cached_dir: create cached directory
        :param invalid_post_processing_cache: invalid post processing cache if cached_dir is not None
        """
        self._df = df
        self._cached_dir = cached_dir
        self._invalid_post_processing_cache = invalid_post_processing_cache

        for field in self._required_fields:
            if field not in df.columns:
                raise RuntimeError(f'field not found: {field} -> {df.columns}')

        self.__allow_inplace = True

    def __repr__(self):
        return repr(self.dataframe())

    def dataframe(self, dataframe: pl.DataFrame = None, may_inplace=True):
        if dataframe is None:
            return self._df
        else:
            return RoiClassifierDataFrame(dataframe, cached_dir=self._cached_dir)

    @property
    def channels(self) -> list[Channel]:
        """list of channel names"""
        return self.dataframe()['channel'].unique().to_list()

    @property
    def n_channels(self) -> int:
        """number of channel"""
        return len(self.channels)

    @property
    def channel_counts(self) -> pl.DataFrame:
        """channel counts dataframe"""
        return self['channel'].value_counts()

    @property
    def source_counts(self) -> pl.DataFrame:
        """source counts dataframe"""
        return self['source'].value_counts()

    @property
    def sources(self) -> list[Channel]:
        """list of source names"""
        return self.dataframe()['source'].unique().to_list()

    @property
    def n_sources(self) -> int:
        """number of source"""
        return len(self.sources)

    @property
    def is_overlapped_channel(self) -> bool:
        """whether there is overlapped channel"""
        return 'overlap' in self.channels

    def get_channel_source_dict(self) -> dict[str, str]:
        """get channel (key): source (value) dict"""
        return {
            it[0]: it[1]
            for it in self['channel', 'source'].unique().iter_rows()
        }

    def get_classified_column(self, field: int | str | None, strict: bool = True) -> str:
        """classified column name, with the given

        :param field: int: ``tree_[level]``; str: ``get column``; None: ``acronym``
        :param strict: strict check for the pre-defined classified column
        """
        match field:
            case int():
                col = f'tree_{field}'
            case str():
                col = field
            case None:
                col = 'acronym'
            case _:
                raise TypeError(f'{type(field)}')

        if strict:
            if col not in self._valid_classified_fields:
                raise ValueError(f'invalid field: {field}')

        if col not in self.dataframe().columns:
            raise ColumnNotFoundError(f'{col} not found: {self.dataframe().columns}')

        return col

    def post_processing(self, *,
                        filter_capital: bool = True,
                        tree: bool = True,
                        family: bool = True,
                        hemisphere: bool = True,
                        copy_overlap: bool = True,
                        filter_injection: tuple[str, str] | None = None) -> Self:
        """
        Load the post-processing dataframe

        :param filter_capital: filter only ``acronym`` contain capital letters
        :param tree: with customized hierarchical tree structure based on allen brain
        :param family: with column ``family`` ('HB', 'HY', 'TH', 'MB', 'CB', 'CTXpl', 'HPF', 'ISOCORTEX', 'OLF', 'CTXsp'
        :param hemisphere: with column ``hemisphere`` with which hemisphere
        :param copy_overlap: Copy overlap channels counts to individual channels, only set true if overlapped roi is not counted for individual channels
        :param filter_injection:  filter out the injection site labelled
        :return:
        """
        if self._cached_dir is not None:
            file = ensure_dir(self._cached_dir) / 'parsed_roi.csv'

            if file.exists() and self._invalid_post_processing_cache:
                file.unlink()

            # load
            if file.exists():
                df = pl.read_csv(file)
                print_load(file)
                return RoiClassifierDataFrame(
                    df, cached_dir=self._cached_dir,
                    invalid_post_processing_cache=self._invalid_post_processing_cache
                )

            # write
            ret = self._post_processing(filter_capital, tree, family, hemisphere, copy_overlap, filter_injection)
            ret.dataframe().write_csv(file)
            print_save(file)
            return ret

        # return
        return self._post_processing(filter_capital, tree, family, hemisphere, copy_overlap, filter_injection)

    def _post_processing(self, filter_capital, tree, family, hemisphere, copy_overlap, filter_injection) -> Self:
        ret = self
        if filter_capital:
            ret = ret.filter_capital_name()
        if tree:
            ret = ret.with_tree_columns()
        if family:
            ret = ret.with_family_columns()
        if hemisphere:
            ret = ret.with_hemisphere_column()
        if copy_overlap:
            ret = ret.with_overlap_copy()
            self.__allow_inplace = False
        if filter_injection is not None:
            ret = ret.filter_injection_site(area=filter_injection[0], hemisphere=filter_injection[1])

        return ret

    def filter_injection_site(self, area: str, hemisphere: str) -> Self:
        """
        filter out the injection site labelled

        :param area: brain area
        :param hemisphere: which hemisphere
        :return:
        """
        expr1 = pl.col('acronym').str.starts_with(area)
        expr2 = pl.col('hemisphere') == hemisphere
        return self.filter(~(expr1 & expr2))

    def filter_capital_name(self) -> Self:
        """filter only ``acronym`` contain capital letters"""
        return self.filter(pl.col('acronym').str.contains(r'[A-Z]+'))

    def with_tree_columns(self) -> Self:
        """with customized hierarchical tree structure based on allen brain

        .. seealso::

            Wang et al 2020, https://doi.org/10.1016/j.cell.2020.04.007
        """
        acronym = self['acronym']
        return self.with_columns(
            pl.Series(name=f'tree_{level}', values=merge_until_level(acronym, level))
            for level in range(NUM_MERGE_LAYER)
        )

    def with_family_columns(self) -> Self:
        """with column ``family`` ('HB', 'HY', 'TH', 'MB', 'CB', 'CTXpl', 'HPF', 'ISOCORTEX', 'OLF', 'CTXsp')"""

        def get_family(row) -> str:
            for name, family in DEFAULT_FAMILY_DICT.items():
                if row in family:
                    return name
            return 'unknown'

        return self.with_columns(pl.col('tree_0').map_elements(get_family, return_dtype=pl.Utf8).alias('family'))

    def with_hemisphere_column(self, invert: bool = False) -> Self:
        """
        with column ``hemisphere`` with which hemisphere

        :param invert: invert hemisphere. Default ML >= 0 (ipsi), ML < 0 (contra)
        :return:
        """
        if invert:
            expr = pl.when(pl.col('ML_location') < 0)
        else:
            expr = pl.when(pl.col('ML_location') >= 0)

        return self.with_columns(expr.then(pl.lit('ipsi')).otherwise(pl.lit('contra')).alias('hemisphere'))

    def with_overlap_copy(self) -> Self:
        """Copy overlap channels counts to individual channels, only used if overlapped roi is not counted for individual channels"""
        if not self.__allow_inplace:
            raise RuntimeError('recurrent copy overlap')

        ret = [self.dataframe()]
        for channel, source in self.get_channel_source_dict().items():
            if channel not in 'overlap':
                df = (
                    self.dataframe().filter(pl.col('channel') == 'overlap')
                    .with_columns(pl.lit(channel).alias('channel'))
                    .with_columns(pl.lit(source).alias('source'))
                )
                ret.append(df)

        return RoiClassifierDataFrame(pl.concat(ret), cached_dir=self._cached_dir)

    # ==================== #
    # Normalized DataFrame #
    # ==================== #

    def to_normalized(self, norm: ROIS_NORM_TYPE,
                      level: int | str | None, *,
                      source: str | None = None,
                      top_area: int | None = None,
                      rest_as_others: bool = False,
                      hemisphere: HEMISPHERE_TYPE = 'both',
                      animal: str | None = None,
                      volume_norm_backend: Literal['cellatlas', 'brainglobe'] = 'cellatlas') -> RoiNormalizedDataFrame:
        """
        To the normalized dataframe (example as volume normalized) ::

            ┌─────────┬────────┬────────┬───────────┬────────────┬────────────────┬────────────┐
            │ source  ┆ tree_2 ┆ counts ┆ fraction  ┆ hemisphere ┆ Volumes [mm^3] ┆ normalized │
            │ ---     ┆ ---    ┆ ---    ┆ ---       ┆ ---        ┆ ---            ┆ ---        │
            │ str     ┆ str    ┆ u32    ┆ f64       ┆ str        ┆ f64            ┆ f64        │
            ╞═════════╪════════╪════════╪═══════════╪════════════╪════════════════╪════════════╡
            │ overlap ┆ ACA    ┆ 1208   ┆ 29.997517 ┆ both       ┆ 5.222484       ┆ 231.307537 │
            │ pRSC    ┆ ACA    ┆ 3296   ┆ 22.822324 ┆ both       ┆ 5.222484       ┆ 631.117254 │
            │ …       ┆ …      ┆ …      ┆ …         ┆ …          ┆ …              ┆ …          │
            │ pRSC    ┆ VIS    ┆ 4035   ┆ 27.939344 ┆ both       ┆ 12.957203      ┆ 311.409797 │
            │ overlap ┆ VIS    ┆ 628    ┆ 15.594736 ┆ both       ┆ 12.957203      ┆ 48.46725   │
            │ aRSC    ┆ VIS    ┆ 3865   ┆ 12.627005 ┆ both       ┆ 12.957203      ┆ 298.289682 │
            └─────────┴────────┴────────┴───────────┴────────────┴────────────────┴────────────┘


        :param norm: :attr:`~neuralib.atlas.ccf.classifier.ROIS_NORM_TYPE`
        :param level: tree level for determine which level of classified column
        :param source: filter only the given ``source``
        :param top_area: filter only the given top areas (sorted based on ``fraction``)
        :param rest_as_others: vertical concat a region called **other** with the roi filtered by ``top area``
        :param hemisphere:  filter only the given ``hemisphere``
        :param animal: with animal id column ``animal``
        :param volume_norm_backend: volume normalization backend. {'cellatlas', 'brainglobe'}
        :return: ``RoiNormalizedDataFrame``
        """
        cls_col = self.get_classified_column(level)

        if hemisphere != 'both':
            cols = ['source', cls_col, 'hemisphere']
        else:
            cols = ['source', cls_col]

        df = (
            self.dataframe().select(cols).group_by(cols).agg(pl.col(cls_col).count().alias('counts'))
            .with_columns((pl.col('counts') / pl.col('counts').sum().over('source') * 100).alias('fraction'))
            .sort('fraction', descending=True)
        )

        #
        if hemisphere != 'both':
            df = df.filter(pl.col('hemisphere') == hemisphere)
        else:
            df = df.with_columns(pl.lit(hemisphere).alias('hemisphere'))  # add back

        #
        if source is not None:
            df = df.filter(pl.col('source') == source)

        #
        if top_area is not None:
            if source is not None:
                df = self._filter_top_region_single_source(df, top_area, source, rest_as_others)
            else:
                df = self._filter_top_region_all_source(df, top_area, level)

        ret = RoiNormalizedDataFrame(df, cls_col, norm)

        #
        match norm:
            case 'volume':
                ret = ret.with_density_column(backend=volume_norm_backend)
            case 'cell':
                ret = ret.with_cell_density_column()
            case 'channel':
                ret = ret.with_columns(pl.col('fraction').alias('normalized'))  # copy from fraction
            case 'none':
                pass
            case _:
                raise ValueError(f'invalid norm method: {norm}')
        #
        if animal is not None:
            ret = ret.with_animal_column(animal)

        return ret

    def _filter_top_region_single_source(self, df, top_area, source, rest_as_others) -> pl.DataFrame:
        if top_area > df.shape[0]:
            print(f'{top_area} areas exceed, thus use all areas instead')
        else:
            df = df[:top_area]

        others = df['fraction'].sum()
        if rest_as_others and (100 - others) > 0:
            other_perc = max(0, 100 - others)
            total_counts = self.source_counts.filter(pl.col('source') == source)['count'].item()
            other_counts = total_counts - df['counts'].sum()

            schema = {df.columns[i]: df.dtypes[i] for i in range(df.shape[1])}
            row = pl.DataFrame([[source, 'other', other_counts, other_perc, df['hemisphere'][0]]], schema=schema,
                               orient='row')  # `other` row

            df = pl.concat([df, row])

        return df

    def _filter_top_region_all_source(self, df, top_area, level) -> pl.DataFrame:
        if top_area > df.shape[0]:
            print(f'{top_area} areas exceed, thus use all areas instead')
            ref_df = df
        else:
            ref_df = df[:top_area]

        cls_col = self.get_classified_column(level)
        region = ref_df[cls_col].unique()

        return df.filter(pl.col(cls_col).is_in(region))

    def to_subregion(self, region: str, *,
                     unit: Literal['counts', 'fraction'] = 'fraction',
                     source_order: tuple[str, ...] | None = None,
                     show_col: str | None = None,
                     animal: str | None = None,
                     normalize: bool = True) -> RoiSubregionDataFrame:
        """
        To the subregion dataframe (example as Visual region: VIS) ::

            ┌─────────┬───────────┬───────────┬───────────┬───┬──────────┬──────────┬──────────┬──────────┐
            │ source  ┆ VISam     ┆ VISp      ┆ VISpm     ┆ … ┆ VISal    ┆ VISpor   ┆ VISli    ┆ VISpl    │
            │ ---     ┆ ---       ┆ ---       ┆ ---       ┆   ┆ ---      ┆ ---      ┆ ---      ┆ ---      │
            │ str     ┆ f64       ┆ f64       ┆ f64       ┆   ┆ f64      ┆ f64      ┆ f64      ┆ f64      │
            ╞═════════╪═══════════╪═══════════╪═══════════╪═══╪══════════╪══════════╪══════════╪══════════╡
            │ overlap ┆ 39.649682 ┆ 15.127389 ┆ 28.025478 ┆ … ┆ 3.025478 ┆ 2.707006 ┆ 1.592357 ┆ 0.159236 │
            │ aRSC    ┆ 32.160414 ┆ 28.952135 ┆ 23.05304  ┆ … ┆ 6.080207 ┆ 1.293661 ┆ 2.069858 ┆ 0.07762  │
            │ pRSC    ┆ 25.947955 ┆ 27.95539  ┆ 27.459727 ┆ … ┆ 3.122677 ┆ 2.973978 ┆ 1.982652 ┆ 1.016109 │
            └─────────┴───────────┴───────────┴───────────┴───┴──────────┴──────────┴──────────┴──────────┘

        :param region: region name
        :param unit: value unit. {'counts', 'fraction'}. default is 'fraction'
        :param source_order: source order in dataframe (rows)
        :param show_col: force set show col to which level. use case: if a low level area name is classified and show in high level (i.e., TH).
        :param animal: with animal column as subregion dataframe
        :param normalize: do subregion normalization. True: normalize per source within region (each source sums to 100% within this region),
            False: normalize per source to total input counts (fraction of total input for that source). Default is True
        :return:
        """

        orig_df = self.dataframe()
        source_order = source_order or tuple(orig_df['source'].unique().to_list())

        # query based on lowest tree level
        query_col = get_margin_merge_level(orig_df, region, 'lowest')

        # show based on highest tree level
        show_col = show_col or get_margin_merge_level(orig_df, region, 'highest')
        s, lv = show_col.rsplit('_', 1)
        next_lv = f'{s}_{int(lv) + 1}'
        show_col = f'{s}_{int(lv)}' if next_lv not in orig_df.columns else next_lv

        df = (
            orig_df
            .filter(pl.col(query_col) == region)
            .select(['source', show_col])
            .group_by(['source', show_col])
            .agg(pl.col(show_col).count().alias('counts'))
        )

        if normalize:
            # normalize per source within this region (each source sums to 100% within the region)
            df = df.with_columns((pl.col('counts') / pl.col('counts').sum().over('source') * 100).alias('fraction'))
        else:
            # normalize per source to total input counts for that source
            source_totals = orig_df.group_by('source').len(name='total_input_counts')
            df = df.join(source_totals, on='source')
            df = df.with_columns((pl.col('counts') / pl.col('total_input_counts') * 100).alias('fraction'))
            df = df.drop('total_input_counts')

        df = df.sort('fraction', descending=True)

        # sort
        idx = {val: idx for idx, val in enumerate(source_order)}
        sort_expr = pl.col('source').replace(idx)

        # profile
        roi_profile = (
            df.group_by('source').agg(pl.col('counts').sum().alias('counts'))
            .join(orig_df.group_by('source').len(name='total'), on='source')
            .with_columns((pl.col('counts') / pl.col('total')).alias('total_fraction'))
            .sort(sort_expr)
        )

        # main result
        ret = (
            df.pivot(show_col, index='source', values=unit, aggregate_function='first')
            .fill_null(0)
            .sort(sort_expr)
        )

        subregion = RoiSubregionDataFrame(region, ret, roi_profile)

        if animal is not None:
            subregion = subregion.with_animal_column(animal)

        return subregion


class RoiNormalizedDataFrame(DataFrameWrapper):
    """
    RoiNormalizedDataFrame with each area per row (unique ``source``, ``hemisphere``)

    - Required fields:
        - ``counts``: roi counts
        - ``fraction``: roi fraction for individual sources (aka. per channel(source) normalized)
        - ``hemisphere``: which hemisphere
        - area column field {'acronym', 'tree_0', 'tree_1', 'tree_2', 'tree_3', 'tree_4'}

    - Optional field:
        - normalization-specific fields (if not 'none'): ``normalized``, ``Volumes [mm^3]``, ``volume_mm3``, ``n_neurons``

    """
    _required_fields = ('counts', 'fraction', 'hemisphere')

    def __init__(self, df: pl.DataFrame,
                 classified_column: str,
                 normalized: ROIS_NORM_TYPE):
        """
        :param df: DataFrame with required fields
        :param classified_column: classified column for the brain area
        :param normalized: :attr:`~neuralib.atlas.ccf.classifier.ROIS_NORM_TYPE`
        """

        for field in self._required_fields:
            if field not in df.columns:
                raise RuntimeError(f'field not found: {field}')

        self._df = df
        self._classified_column = classified_column
        self._normalized = normalized

    def __repr__(self):
        return repr(self.dataframe())

    @property
    def classified_column(self) -> str:
        """region classified column name"""
        return self._classified_column

    @property
    def normalized(self) -> ROIS_NORM_TYPE:
        """normalization type"""
        return self._normalized

    @property
    def value_column(self) -> str:
        """value column based on the ``normalized``"""
        match self._normalized:
            case 'volume' | 'cell' | 'channel':
                return 'normalized'
            case 'none':
                return 'counts'
            case _:
                raise ValueError(f'invalid normalized method: {self._normalized}')

    @property
    def normalized_unit(self) -> str:
        """unit based on the ``normalized``"""
        match self._normalized:
            case 'volume':
                return 'density (cells-mm3)'
            case 'cell':
                return 'cell density (%)'
            case 'channel':
                return 'fraction (%)'
            case 'none':
                return 'counts'
            case _:
                raise ValueError(f'invalid normalized unit: {self._normalized}')

    def dataframe(self, dataframe: pl.DataFrame = None, may_inplace=True):
        """
        RoiNormalizedDataFrame (Volume normalized as example) ::

            ┌─────────┬────────┬────────┬───────────┬────────────┬────────────────┬────────────┐
            │ source  ┆ tree_2 ┆ counts ┆ fraction  ┆ hemisphere ┆ Volumes [mm^3] ┆ normalized │
            │ ---     ┆ ---    ┆ ---    ┆ ---       ┆ ---        ┆ ---            ┆ ---        │
            │ str     ┆ str    ┆ u32    ┆ f64       ┆ str        ┆ f64            ┆ f64        │
            ╞═════════╪════════╪════════╪═══════════╪════════════╪════════════════╪════════════╡
            │ overlap ┆ ACA    ┆ 1208   ┆ 29.997517 ┆ both       ┆ 5.222484       ┆ 231.307537 │
            │ pRSC    ┆ ACA    ┆ 3296   ┆ 22.822324 ┆ both       ┆ 5.222484       ┆ 631.117254 │
            │ …       ┆ …      ┆ …      ┆ …         ┆ …          ┆ …              ┆ …          │
            │ pRSC    ┆ VIS    ┆ 4035   ┆ 27.939344 ┆ both       ┆ 12.957203      ┆ 311.409797 │
            │ overlap ┆ VIS    ┆ 628    ┆ 15.594736 ┆ both       ┆ 12.957203      ┆ 48.46725   │
            │ aRSC    ┆ VIS    ┆ 3865   ┆ 12.627005 ┆ both       ┆ 12.957203      ┆ 298.289682 │
            └─────────┴────────┴────────┴───────────┴────────────┴────────────────┴────────────┘

        """
        if dataframe is None:
            return self._df
        else:
            ret = RoiNormalizedDataFrame(dataframe, self._classified_column, self._normalized)
            return ret

    def with_density_column(self, backend: Literal['cellatlas', 'brainglobe'] = 'cellatas') -> Self:
        """

        :param backend: Volume information calculated from which backend. {'cellatlas', 'brainglobe'}
        :return:
        """
        match backend:
            case 'cellatlas':
                df_cellatlas = load_cellatlas().select('Volumes [mm^3]', 'acronym').rename(
                    {'acronym': self.classified_column})
                return (self.join(df_cellatlas, on=self.classified_column)
                        .with_columns((pl.col('counts') / pl.col('Volumes [mm^3]')).alias('normalized')))

            case 'brainglobe':
                df_bg_volume = load_bg_volumes().select('volume_mm3', 'acronym').rename(
                    {'acronym': self.classified_column})
                return (self.join(df_bg_volume, on=self.classified_column)
                        .with_columns((pl.col('counts') / pl.col('volume_mm3')).alias('normalized')))

            case _:
                raise ValueError(f'invalid backend: {backend}')

    def with_cell_density_column(self) -> Self:
        """Normalized to number of neurons foreach brain region (based on ``CellAtlas`` data source)"""
        df_cellatlas = load_cellatlas().select('n_neurons', 'acronym').rename({'acronym': self.classified_column})
        return (self.join(df_cellatlas, on=self.classified_column)
                .with_columns((pl.col('counts') / pl.col('n_neurons')).alias('normalized')))

    def with_animal_column(self, animal) -> Self:
        """with animal id column"""
        return self.with_columns(pl.lit(animal).alias('animal'))

    def filter_areas(self, areas: str | list[str]) -> Self:
        """filter the dataframe with specified areas"""
        if isinstance(areas, str):
            areas = [areas]

        ret = self.filter(pl.col(self.classified_column).is_in(areas))
        if ret.dataframe().is_empty():
            raise ValueError(f'{areas} not found')

        return ret

    def filter_sources(self, source: str | list[str]) -> Self:
        """filter the dataframe with specified sources"""
        if isinstance(source, str):
            source = [source]

        ret = self.filter(pl.col('source').is_in(source))
        if ret.dataframe().is_empty():
            raise ValueError(f'{source} not found')

        return ret

    def to_bias_index(self, source_a: str, source_b: str) -> pl.DataFrame:
        """
        Bias Index dataframe used to determine bias within two sources
        (positive value toward ``source a`` and negative value toward ``source b``) ::

            ┌────────┬────────────┐
            │ tree_2 ┆ bias_index │
            │ ---    ┆ ---        │
            │ str    ┆ f64        │
            ╞════════╪════════════╡
            │ ATN    ┆ -1.192889  │
            │ VIS    ┆ -1.145786  │
            │ CLA    ┆ -0.86059   │
            │ SUB    ┆ -0.478069  │
            │ STRd   ┆ -0.463589  │
            │ …      ┆ …          │
            │ ENT    ┆ 0.580593   │
            │ AUD    ┆ 0.610688   │
            │ PTLp   ┆ 1.292926   │
            │ MO     ┆ 1.945567   │
            │ SS     ┆ 2.163074   │
            └────────┴────────────┘

        :param source_a: source a string
        :param source_b: source b string
        :return:
        """
        expr_calc = (pl.col(source_a) / pl.col(source_b)).map_elements(np.log2, return_dtype=pl.Float64)
        df = (
            self.dataframe()
            .select(self.classified_column, 'source', 'fraction')
            .sort(self.classified_column, 'source')
            .pivot(values='fraction', index=self.classified_column, on='source', aggregate_function='first')
            .fill_null(0)
            .with_columns(expr_calc.alias('bias_index'))
            .filter(~pl.col('bias_index').is_infinite())  # log2 index calc fail
            .select(self.classified_column, 'bias_index')
            .sort(by='bias_index')
        )

        return df

    def to_winner(self, sources: Sequence[str]) -> pl.DataFrame:
        """
        Winner dataframe used for plotting (i.e., ternary plot) ::

            ┌────────┬─────────┬──────┬──────┬───────┬────────┐
            │ tree_2 ┆ overlap ┆ pRSC ┆ aRSC ┆ total ┆ winner │
            │ ---    ┆ ---     ┆ ---  ┆ ---  ┆ ---   ┆ ---    │
            │ str    ┆ u32     ┆ u32  ┆ u32  ┆ u32   ┆ str    │
            ╞════════╪═════════╪══════╪══════╪═══════╪════════╡
            │ ACA    ┆ 1208    ┆ 3296 ┆ 5761 ┆ 9057  ┆ aRSC   │
            │ VIS    ┆ 628     ┆ 4035 ┆ 3865 ┆ 7900  ┆ pRSC   │
            │ MO     ┆ 460     ┆ 714  ┆ 5829 ┆ 6543  ┆ aRSC   │
            │ …      ┆ …       ┆ …    ┆ …    ┆ …     ┆ …      │
            │ AUD    ┆ 44      ┆ 165  ┆ 534  ┆ 699   ┆ aRSC   │
            │ TEa    ┆ 34      ┆ 206  ┆ 358  ┆ 564   ┆ aRSC   │
            └────────┴─────────┴──────┴──────┴───────┴────────┘

        :param sources: source sequences for calculating the total. The above case should be specified as ['aRSC', 'pRSC']
        :return: Winner dataframe
        """
        df = (
            self.dataframe()
            .pivot(values=self.value_column, on='source', index=self.classified_column, aggregate_function='first')
            .fill_nan(0)
            .fill_null(0)
            .with_columns(pl.sum_horizontal(sources).alias('total'))
        )

        region_counts = df.select(sources).to_numpy()
        winner_idx = np.argmax(region_counts, axis=1).astype(int)
        df = df.with_columns(pl.Series([sources[idx] for idx in winner_idx]).alias('winner'))

        return df


class RoiSubregionDataFrame(DataFrameWrapper):
    """RoiSubregionDataFrame with each source per row, column shows the subregions"""

    _profile_required_fields = ('source', 'counts', 'total', 'total_fraction')

    def __init__(self, region: str, df: pl.DataFrame, profile: pl.DataFrame):
        """
        :param region: region name
        :param df: subregion dataframe
        :param profile: profile dataframe
        """
        self._region = region
        self._df = df
        self._profile = profile

        for field in self._profile_required_fields:
            if field not in profile.columns:
                raise RuntimeError(f'field not found: {field} -> {df.columns}')

    def __repr__(self):
        return repr(self.dataframe())

    def dataframe(self, dataframe: pl.DataFrame = None, may_inplace=True):
        """
        RoiSubregionDataFrame (VIS as example)::

            ┌─────────┬───────────┬───────────┬───────────┬───┬──────────┬──────────┬──────────┬──────────┐
            │ source  ┆ VISam     ┆ VISp      ┆ VISpm     ┆ … ┆ VISal    ┆ VISpor   ┆ VISli    ┆ VISpl    │
            │ ---     ┆ ---       ┆ ---       ┆ ---       ┆   ┆ ---      ┆ ---      ┆ ---      ┆ ---      │
            │ str     ┆ f64       ┆ f64       ┆ f64       ┆   ┆ f64      ┆ f64      ┆ f64      ┆ f64      │
            ╞═════════╪═══════════╪═══════════╪═══════════╪═══╪══════════╪══════════╪══════════╪══════════╡
            │ overlap ┆ 39.649682 ┆ 15.127389 ┆ 28.025478 ┆ … ┆ 3.025478 ┆ 2.707006 ┆ 1.592357 ┆ 0.159236 │
            │ aRSC    ┆ 32.160414 ┆ 28.952135 ┆ 23.05304  ┆ … ┆ 6.080207 ┆ 1.293661 ┆ 2.069858 ┆ 0.07762  │
            │ pRSC    ┆ 25.947955 ┆ 27.95539  ┆ 27.459727 ┆ … ┆ 3.122677 ┆ 2.973978 ┆ 1.982652 ┆ 1.016109 │
            └─────────┴───────────┴───────────┴───────────┴───┴──────────┴──────────┴──────────┴──────────┘
        """
        if dataframe is None:
            return self._df
        else:
            ret = RoiSubregionDataFrame(self._region, dataframe, self._profile)
            return ret

    @property
    def region(self) -> str:
        """region name"""
        return self._region

    @property
    def subregion(self) -> list[str]:
        """list of subregion names"""
        return self.drop('source').columns

    @property
    def n_subregion(self) -> int:
        """number of subregion"""
        return len(self.subregion)

    @property
    def profile(self) -> pl.DataFrame:
        """
        with channel-wise profile::

            ┌─────────┬────────┬───────┬────────────────┐
            │ source  ┆ counts ┆ total ┆ total_fraction │
            │ ---     ┆ ---    ┆ ---   ┆ ---            │
            │ str     ┆ u32    ┆ u32   ┆ f64            │
            ╞═════════╪════════╪═══════╪════════════════╡
            │ overlap ┆ 628    ┆ 4027  ┆ 0.155947       │
            │ aRSC    ┆ 3865   ┆ 30609 ┆ 0.12627        │
            │ pRSC    ┆ 4035   ┆ 14442 ┆ 0.279393       │
            └─────────┴────────┴───────┴────────────────┘
        """
        return self._profile

    @property
    def sources(self) -> list[str]:
        """list of source names"""
        return self['source'].to_list()

    def filter_overlap(self) -> Self:
        """filter out overlap source"""
        expr = pl.col('source') != 'overlap'
        self._profile = self._profile.filter(expr)
        return self.filter(expr)

    def with_animal_column(self, animal) -> Self:
        """with animal id column"""
        return self.with_columns(pl.lit(animal).alias('animal'))

    def to_dict(self, as_series: bool = True) -> dict[str, list[float]]:
        """to subregion:value dict"""
        return self.dataframe().select(pl.exclude('source')).to_dict(as_series=as_series)

    def to_numpy(self) -> np.ndarray:
        """to value array. `Array[float, [n_source, n_subregion]]`"""
        return self.dataframe().select(pl.exclude('source')).to_numpy()



from __future__ import annotations

from os import PathLike
from typing import TypedDict, Any

import attrs
import numpy as np
from typing_extensions import Self

from neuralib.util.verbose import print_load, print_save

__all__ = ['read_rastermap',
           'save_rastermap',
           'RasterMapResult',
           'UserCluster',
           'RasterOptions']


def read_rastermap(file: PathLike) -> RasterMapResult:
    """load rastermap result
    :param file: output from rastermap. *_embedding.npy
    :return:
    """
    return RasterMapResult.load(file)


def save_rastermap(result: RasterMapResult, path: PathLike):
    """save result for GUI relaunch

    :param result: :class:`~RasterMapResult`
    :param path: output from rastermap. *_embedding.npy
    """
    result.save(path)


@attrs.define
class RasterMapResult:
    """Container for storing the rastermap result,
    For both GUI load and customized plotting purpose

    `Dimension parameters`:

        N = Number of neurons/pixel

        T = Number of samples

        C = Number of clusters = N / binsize

    """

    filename: str
    """Filename of the neural activity data
    (i.e., *.tif or *.avi for wfield activity; .npy `Array[float, [N, T]]` file for cellular)"""

    save_path: str
    """filename for the rastermap result save"""

    isort: np.ndarray
    """`Array[int, N]`"""

    embedding: np.ndarray
    """`Array[float, [N, 1]]`"""

    ops: RasterOptions
    """``RasterOptions``"""

    user_clusters: list[UserCluster] = attrs.field(default=attrs.Factory(list))
    """list of clusters ``UserCluster``"""

    super_neurons: np.ndarray | None = attrs.field(default=None)
    """super neuron activity. `Array[float, [C, T]]`"""

    @classmethod
    def load(cls, path: PathLike) -> Self:
        """
        Load the results from rastermap output

        :param path: file path of the rastermap output
        :return: :class:`RasterMapResult`
        """
        dat = np.load(path, allow_pickle=True).item()
        print_load(path)

        return RasterMapResult(
            filename=dat.get('filename', None),
            save_path=dat.get('save_path', None),
            isort=dat.get('isort', None),
            embedding=dat.get('embedding', None),
            ops=dat.get('ops', None),
            user_clusters=dat.get('user_clusters', None),
            super_neurons=dat.get('super_neurons', None)
        )

    def save(self, path: PathLike) -> None:
        """For GUI loading & cache computing for plotting in different time domains"""
        proc = {
            'filename': self.filename,
            'save_path': self.save_path,
            'isort': self.isort,
            'embedding': self.embedding,
            'ops': self.ops,
            'user_clusters': self.user_clusters,
            'super_neurons': self.super_neurons,
        }

        np.save(path, proc, allow_pickle=True)
        print_save(path)

    def asdict(self) -> dict[str, Any]:
        """result as dict"""
        return attrs.asdict(self)

    @property
    def n_clusters(self) -> int:
        """number of clusters (super neurons)"""
        try:
            return self.super_neurons.shape[0]
        except AttributeError:
            raise RuntimeError('data incomplete')

    @property
    def n_samples(self) -> int:
        """number of data samples (T)"""
        try:
            return self.super_neurons.shape[1]
        except AttributeError:
            raise RuntimeError('data incomplete')


class UserCluster(TypedDict, total=False):
    """GUI selected clusters"""
    ids: np.ndarray
    """Neuronal ids. `Array[int, N]`"""
    slice: slice
    """Binned neurons range"""
    binsize: int
    """Neuron bins"""
    color: np.ndarray
    """Colors. `Array[float, 4]`"""


class RasterOptions(TypedDict, total=False):
    """Run Rastermap model options. Refer to the ``rastermap.rastermap.setting_info()``"""

    n_clusters: int
    """Number of clusters created from data before upsampling and creating embedding 
    (any number above 150 will be very slow due to NP-hard sorting problem)"""

    n_PCs: int
    """Number of PCs to use during optimization"""

    time_lag_window: float
    """Number of time points into the future to compute cross-correlation, useful for sequence finding"""

    locality: float
    """How local should the algorithm be -- set to 1.0 for highly local + sequence finding"""

    n_splits: int
    """Recluster and sort n_splits times (increases local neighborhood preservation)"""

    time_bin: int
    """Binning of data in time before PCA is computed"""

    grid_upsample: int
    """How much to upsample clusters"""

    mean_time: bool
    """Whether to project out the mean over data samples at each timepoint, usually good to keep on to find structure"""

    verbose: bool
    """Whether to output progress during optimization"""

    verbose_sorting: bool
    """Output progress in travelling salesman"""

    start_time: int
    """Optional start time"""

    end_time: int
    """Optional end time"""
